"""FInite element Automatic Tabulator -- supports constructing and
evaluating arbitrary order Lagrange and many other elements.
Simplices in one, two, and three dimensions are supported."""

# Important functionality
from FIAT.reference_element import ufc_cell, ufc_simplex       # noqa: F401
from FIAT.quadrature import make_quadrature                    # noqa: F401
from FIAT.quadrature_schemes import create_quadrature          # noqa: F401
from FIAT.finite_element import FiniteElement, CiarletElement  # noqa: F401
from FIAT.hdivcurl import Hdiv, Hcurl                          # noqa: F401
from FIAT.mixed import MixedElement                            # noqa: F401
from FIAT.restricted import RestrictedElement                  # noqa: F401
from FIAT.quadrature_element import QuadratureElement          # noqa: F401
from FIAT.tensor_product import TensorProductElement           # noqa: F401
from FIAT.enriched import EnrichedElement                      # noqa: F401
from FIAT.nodal_enriched import NodalEnrichedElement           # noqa: F401
from FIAT.discontinuous import DiscontinuousElement            # noqa: F401

# Import finite element classes
from FIAT.argyris import Argyris
from FIAT.bernardi_raugel import BernardiRaugel
from FIAT.bernstein import Bernstein
from FIAT.bell import Bell
from FIAT.hct import HsiehCloughTocher
from FIAT.alfeld_sorokina import AlfeldSorokina
from FIAT.arnold_qin import ArnoldQin
from FIAT.guzman_neilan import GuzmanNeilanFirstKindH1, GuzmanNeilanSecondKindH1, GuzmanNeilanH1div
from FIAT.christiansen_hu import ChristiansenHu
from FIAT.johnson_mercier import JohnsonMercier
from FIAT.brezzi_douglas_marini import BrezziDouglasMarini
from FIAT.Sminus import TrimmedSerendipityEdge, TrimmedSerendipityFace
from FIAT.SminusDiv import TrimmedSerendipityDiv
from FIAT.SminusCurl import TrimmedSerendipityCurl
from FIAT.brezzi_douglas_fortin_marini import BrezziDouglasFortinMarini
from FIAT.discontinuous_lagrange import DiscontinuousLagrange
from FIAT.discontinuous_taylor import DiscontinuousTaylor
from FIAT.discontinuous_raviart_thomas import DiscontinuousRaviartThomas
from FIAT.serendipity import Serendipity
from FIAT.brezzi_douglas_marini_cube import BrezziDouglasMariniCubeEdge, BrezziDouglasMariniCubeFace
from FIAT.discontinuous_pc import DPC
from FIAT.hermite import CubicHermite
from FIAT.lagrange import Lagrange
from FIAT.gauss_lobatto_legendre import GaussLobattoLegendre
from FIAT.gauss_legendre import GaussLegendre
from FIAT.gauss_radau import GaussRadau
from FIAT.morley import Morley
from FIAT.nedelec import Nedelec
from FIAT.nedelec_second_kind import NedelecSecondKind
from FIAT.powell_sabin import QuadraticPowellSabin6, QuadraticPowellSabin12
from FIAT.hierarchical import Legendre, IntegratedLegendre
from FIAT.P0 import P0
from FIAT.raviart_thomas import RaviartThomas
from FIAT.crouzeix_raviart import CrouzeixRaviart
from FIAT.regge import Regge
from FIAT.gopalakrishnan_lederer_schoberl import GopalakrishnanLedererSchoberlFirstKind
from FIAT.gopalakrishnan_lederer_schoberl import GopalakrishnanLedererSchoberlSecondKind
from FIAT.hellan_herrmann_johnson import HellanHerrmannJohnson
from FIAT.arnold_winther import ArnoldWinther
from FIAT.arnold_winther import ArnoldWintherNC
from FIAT.hu_zhang import HuZhang
from FIAT.mardal_tai_winther import MardalTaiWinther
from FIAT.bubble import Bubble, FacetBubble
from FIAT.hdiv_trace import HDivTrace
from FIAT.kong_mulder_veldhuizen import KongMulderVeldhuizen
from FIAT.histopolation import Histopolation
from FIAT.fdm_element import FDMLagrange, FDMDiscontinuousLagrange, FDMQuadrature, FDMBrokenH1, FDMBrokenL2, FDMHermite  # noqa: F401

# List of supported elements and mapping to element classes
supported_elements = {"Argyris": Argyris,
                      "Bell": Bell,
                      "Bernardi-Raugel": BernardiRaugel,
                      "Bernstein": Bernstein,
                      "Brezzi-Douglas-Marini": BrezziDouglasMarini,
                      "Brezzi-Douglas-Fortin-Marini": BrezziDouglasFortinMarini,
                      "Bubble": Bubble,
                      "FacetBubble": FacetBubble,
                      "Crouzeix-Raviart": CrouzeixRaviart,
                      "Discontinuous Lagrange": DiscontinuousLagrange,
                      "S": Serendipity,
                      "SminusF": TrimmedSerendipityFace,
                      "SminusDiv": TrimmedSerendipityDiv,
                      "SminusE": TrimmedSerendipityEdge,
                      "SminusCurl": TrimmedSerendipityCurl,
                      "Brezzi-Douglas-Marini Cube Face": BrezziDouglasMariniCubeFace,
                      "Brezzi-Douglas-Marini Cube Edge": BrezziDouglasMariniCubeEdge,
                      "DPC": DPC,
                      "Discontinuous Taylor": DiscontinuousTaylor,
                      "Discontinuous Raviart-Thomas": DiscontinuousRaviartThomas,
                      "Hermite": CubicHermite,
                      "Hsieh-Clough-Tocher": HsiehCloughTocher,
                      "QuadraticPowellSabin6": QuadraticPowellSabin6,
                      "QuadraticPowellSabin12": QuadraticPowellSabin12,
                      "Alfeld-Sorokina": AlfeldSorokina,
                      "Arnold-Qin": ArnoldQin,
                      "Christiansen-Hu": ChristiansenHu,
                      "Guzman-Neilan 1st kind H1": GuzmanNeilanFirstKindH1,
                      "Guzman-Neilan 2nd kind H1": GuzmanNeilanSecondKindH1,
                      "Guzman-Neilan H1(div)": GuzmanNeilanH1div,
                      "Johnson-Mercier": JohnsonMercier,
                      "Lagrange": Lagrange,
                      "Kong-Mulder-Veldhuizen": KongMulderVeldhuizen,
                      "Gauss-Lobatto-Legendre": GaussLobattoLegendre,
                      "Gauss-Legendre": GaussLegendre,
                      "Gauss-Radau": GaussRadau,
                      "Histopolation": Histopolation,
                      "Legendre": Legendre,
                      "Integrated Legendre": IntegratedLegendre,
                      "Morley": Morley,
                      "Nedelec 1st kind H(curl)": Nedelec,
                      "Nedelec 2nd kind H(curl)": NedelecSecondKind,
                      "Raviart-Thomas": RaviartThomas,
                      "Regge": Regge,
                      "HDiv Trace": HDivTrace,
                      "Hellan-Herrmann-Johnson": HellanHerrmannJohnson,
                      "Gopalakrishnan-Lederer-Schoberl 1st kind": GopalakrishnanLedererSchoberlFirstKind,
                      "Gopalakrishnan-Lederer-Schoberl 2nd kind": GopalakrishnanLedererSchoberlSecondKind,
                      "Conforming Arnold-Winther": ArnoldWinther,
                      "Nonconforming Arnold-Winther": ArnoldWintherNC,
                      "Hu-Zhang": HuZhang,
                      "Mardal-Tai-Winther": MardalTaiWinther}

# List of extra elements
extra_elements = {"P0": P0}



# Copyright (C) 2020 Robert C Kirby (Baylor University)
#
# This file is part of FIAT (https://www.fenicsproject.org)
#
# SPDX-License-Identifier:    LGPL-3.0-or-later

import pytest
import numpy

from FIAT import (
    BrezziDouglasMarini, Morley, Argyris, CubicHermite)

from FIAT.reference_element import (
    UFCTriangle,
    make_lattice)

from FIAT.pointwise_dual import compute_pointwise_dual as cpd

T = UFCTriangle()


@pytest.mark.parametrize("element",
                         [CubicHermite(T),
                          Morley(T),
                          Argyris(T),
                          BrezziDouglasMarini(T, 1, variant="integral")])
def test_pw_dual(element):
    deg = element.degree()
    ref_el = element.ref_el
    poly_set = element.poly_set
    pts = make_lattice(ref_el.vertices, deg)

    assert numpy.allclose(element.dual.to_riesz(poly_set),
                          cpd(element, pts).to_riesz(poly_set))



# Copyright (C) 2021 Pablo D. Brubeck
#
# This file is part of FIAT (https://www.fenicsproject.org)
#
# SPDX-License-Identifier:    LGPL-3.0-or-later
#
# Written by Pablo D. Brubeck (brubeck@protonmail.com), 2021

import numpy
from FIAT import reference_element, expansions, polynomial_set
from FIAT.functional import index_iterator


def get_lagrange_points(nodes):
    """Extract singleton point for each node."""
    points = []
    for node in nodes:
        pt, = node.get_point_dict()
        points.append(pt)
    return points


def barycentric_interpolation(nodes, wts, dmat, pts, order=0):
    """Evaluates a Lagrange basis on a line reference element
    via the second barycentric interpolation formula. See Berrut and Trefethen (2004)
    https://doi.org/10.1137/S0036144502417715 Eq. (4.2) & (9.4)
    """
    if pts.dtype == object:
        from sympy import simplify
        sp_simplify = numpy.vectorize(simplify)
    else:
        sp_simplify = lambda x: x
    phi = numpy.add.outer(-nodes, pts.flatten())
    with numpy.errstate(divide='ignore', invalid='ignore'):
        numpy.reciprocal(phi, out=phi)
        numpy.multiply(phi, wts[:, None], out=phi)
        numpy.multiply(1.0 / numpy.sum(phi, axis=0), phi, out=phi)
    phi[phi != phi] = 1.0
    phi = phi.reshape(-1, *pts.shape[:-1])

    phi = sp_simplify(phi)
    results = {(0,): phi}
    for r in range(1, order+1):
        phi = sp_simplify(numpy.dot(dmat, phi))
        results[(r,)] = phi
    return results


def make_dmat(x):
    """Returns Lagrange differentiation matrix and barycentric weights
    associated with x[j]."""
    dmat = numpy.add.outer(-x, x)
    numpy.fill_diagonal(dmat, 1.0)
    wts = numpy.prod(dmat, axis=0)
    numpy.reciprocal(wts, out=wts)
    numpy.divide(numpy.divide.outer(wts, wts), dmat, out=dmat)
    numpy.fill_diagonal(dmat, dmat.diagonal() - numpy.sum(dmat, axis=0))
    return dmat, wts


class LagrangeLineExpansionSet(expansions.LineExpansionSet):
    """Lagrange polynomial expansion set for given points the line."""
    def __init__(self, ref_el, pts):
        self.points = pts
        self.x = numpy.array(pts, dtype="d").flatten()
        self.cell_node_map = expansions.compute_cell_point_map(ref_el, pts, unique=False)
        self.dmats = [None for _ in self.cell_node_map]
        self.weights = [None for _ in self.cell_node_map]
        self.nodes = [None for _ in self.cell_node_map]
        for cell, ibfs in self.cell_node_map.items():
            self.nodes[cell] = self.x[ibfs]
            self.dmats[cell], self.weights[cell] = make_dmat(self.nodes[cell])

        self.degree = max(len(wts) for wts in self.weights)-1
        self.recurrence_order = self.degree + 1
        super().__init__(ref_el)
        self.continuity = None if len(self.x) == sum(len(xk) for xk in self.nodes) else "C0"

    def get_num_members(self, n):
        return len(self.points)

    def get_cell_node_map(self, n):
        return self.cell_node_map

    def get_points(self):
        return self.points

    def get_dmats(self, degree, cell=0):
        return [self.dmats[cell].T]

    def _tabulate_on_cell(self, n, pts, order=0, cell=0, direction=None):
        return barycentric_interpolation(self.nodes[cell], self.weights[cell], self.dmats[cell], pts, order=order)


class LagrangePolynomialSet(polynomial_set.PolynomialSet):

    def __init__(self, ref_el, pts, shape=tuple()):
        if ref_el.get_shape() != reference_element.LINE:
            raise ValueError("Invalid reference element type.")

        expansion_set = LagrangeLineExpansionSet(ref_el, pts)
        degree = expansion_set.degree
        if shape == tuple():
            num_components = 1
        else:
            flat_shape = numpy.ravel(shape)
            num_components = numpy.prod(flat_shape)
        num_exp_functions = expansion_set.get_num_members(degree)
        num_members = num_components * num_exp_functions
        embedded_degree = degree

        # set up coefficients
        if shape == tuple():
            coeffs = numpy.eye(num_members, dtype="d")
        else:
            coeffs_shape = (num_members, *shape, num_exp_functions)
            coeffs = numpy.zeros(coeffs_shape, "d")
            # use functional's index_iterator function
            cur_bf = 0
            for idx in index_iterator(shape):
                for exp_bf in range(num_exp_functions):
                    cur_idx = (cur_bf, *idx, exp_bf)
                    coeffs[cur_idx] = 1.0
                    cur_bf += 1

        super().__init__(ref_el, degree, embedded_degree, expansion_set, coeffs)



# Copyright (C) 2013 Andrew T. T. McRae (Imperial College London)
# Copyright (C) 2015 Jan Blechta
# Copyright (C) 2018 Patrick E. Farrell
#
# This file is part of FIAT (https://www.fenicsproject.org)
#
# SPDX-License-Identifier:    LGPL-3.0-or-later

from FIAT.lagrange import Lagrange
from FIAT.hierarchical import IntegratedLegendre
from FIAT.restricted import RestrictedElement
from itertools import chain


class CodimBubble(RestrictedElement):
    """Bubbles of a certain codimension."""

    def __init__(self, ref_el, degree, codim, variant=None):
        if variant and variant.startswith("integral"):
            CG = IntegratedLegendre
        else:
            CG = Lagrange
        element = CG(ref_el, degree, variant=variant)

        cell_dim = ref_el.get_dimension()
        assert cell_dim == max(element.entity_dofs().keys())
        dofs = list(sorted(chain(*element.entity_dofs()[cell_dim - codim].values())))
        if len(dofs) == 0:
            raise RuntimeError('Bubble element of degree %d and codimension %d has no dofs' % (degree, codim))

        super().__init__(element, indices=dofs)


class Bubble(CodimBubble):
    """The bubble finite element: the dofs of the Lagrange FE in the interior of the cell"""

    def __init__(self, ref_el, degree, variant=None):
        super().__init__(ref_el, degree, codim=0, variant=variant)


class FacetBubble(CodimBubble):
    """The facet bubble finite element: the dofs of the Lagrange FE in the interior of the facets"""

    def __init__(self, ref_el, degree, variant=None):
        super().__init__(ref_el, degree, codim=1, variant=variant)



"""This module defines the UFL finite element classes."""

# Copyright (C) 2008-2016 Martin Sandve Alnæs
#
# This file was originally part of UFL (https://www.fenicsproject.org)
#
# SPDX-License-Identifier:    LGPL-3.0-or-later
#
# Modified by Kristian B. Oelgaard
# Modified by Marie E. Rognes 2010, 2012
# Modified by Massimiliano Leoni, 2016
# Modified by Matthew Scroggs, 2023

from itertools import chain

from ufl.cell import TensorProductCell, as_cell
from finat.ufl.finiteelementbase import FiniteElementBase
from ufl.sobolevspace import DirectionalSobolevSpace


class TensorProductElement(FiniteElementBase):
    r"""The tensor product of :math:`d` element spaces.

    .. math:: V = V_1 \otimes V_2 \otimes ...  \otimes V_d

    Given bases :math:`\{\phi_{j_i}\}` of the spaces :math:`V_i` for :math:`i = 1, ...., d`,
    :math:`\{ \phi_{j_1} \otimes \phi_{j_2} \otimes \cdots \otimes \phi_{j_d}
    \}` forms a basis for :math:`V`.
    """
    __slots__ = ("_sub_elements", "_cell")

    def __init__(self, *elements, **kwargs):
        """Create TensorProductElement from a given list of elements."""
        if not elements:
            raise ValueError("Cannot create TensorProductElement from empty list.")

        keywords = list(kwargs.keys())
        if keywords and keywords != ["cell"]:
            raise ValueError("TensorProductElement got an unexpected keyword argument '%s'" % keywords[0])
        cell = kwargs.get("cell")

        family = "TensorProductElement"

        if cell is None:
            # Define cell as the product of each elements cell
            cell = TensorProductCell(*[e.cell for e in elements])
        else:
            cell = as_cell(cell)

        # Define polynomial degree as a tuple of sub-degrees
        degree = tuple(e.degree() for e in elements)

        # No quadrature scheme defined
        quad_scheme = None

        # match FIAT implementation
        reference_value_shape = tuple(chain(*[e.reference_value_shape for e in elements]))
        if len(reference_value_shape) > 1:
            raise ValueError("Product of vector-valued elements not supported")

        FiniteElementBase.__init__(self, family, cell, degree,
                                   quad_scheme, reference_value_shape)
        self._sub_elements = elements
        self._cell = cell

    def __repr__(self):
        """Doc."""
        return "TensorProductElement(" + ", ".join(repr(e) for e in self._sub_elements) + f", cell={repr(self._cell)})"

    def mapping(self):
        """Doc."""
        if all(e.mapping() == "identity" for e in self._sub_elements):
            return "identity"
        elif all(e.mapping() == "L2 Piola" for e in self._sub_elements):
            return "L2 Piola"
        else:
            return "undefined"

    @property
    def sobolev_space(self):
        """Return the underlying Sobolev space of the TensorProductElement."""
        elements = self._sub_elements
        if all(e.sobolev_space == elements[0].sobolev_space
               for e in elements):
            return elements[0].sobolev_space
        else:
            # Generate a DirectionalSobolevSpace which contains
            # continuity information parametrized by spatial index
            orders = []
            for e in elements:
                # TODO: is this the right value for e_dim
                e_dim = e.cell.topological_dimension()
                e_order = (e.sobolev_space._order,) * e_dim
                orders.extend(e_order)
            return DirectionalSobolevSpace(orders)

    @property
    def num_sub_elements(self):
        """Return number of subelements."""
        return len(self._sub_elements)

    @property
    def sub_elements(self):
        """Return subelements (factors)."""
        return self._sub_elements

    def reconstruct(self, **kwargs):
        """Doc."""
        cell = kwargs.pop("cell", self.cell)
        return TensorProductElement(*[e.reconstruct(**kwargs) for e in self.sub_elements], cell=cell)

    def variant(self):
        """Doc."""
        try:
            variant, = {e.variant() for e in self.sub_elements}
            return variant
        except ValueError:
            return None

    def __str__(self):
        """Pretty-print."""
        return "TensorProductElement(%s, cell=%s)" \
            % (', '.join([str(e) for e in self._sub_elements]), str(self._cell))

    def shortstr(self):
        """Short pretty-print."""
        return "TensorProductElement(%s, cell=%s)" \
            % (', '.join([e.shortstr() for e in self._sub_elements]), str(self._cell))

    @property
    def embedded_superdegree(self):
        """Doc."""
        return sum(self.degree())

    @property
    def embedded_subdegree(self):
        """Doc."""
        return min(self.degree())



import math
import numpy as np
import pytest

from FIAT.reference_element import UFCInterval, UFCTriangle, UFCTetrahedron
from FIAT import create_quadrature, make_quadrature, polynomial_set
from FIAT.kong_mulder_veldhuizen import KongMulderVeldhuizen as KMV

I = UFCInterval()
T = UFCTriangle()
Te = UFCTetrahedron()


@pytest.mark.parametrize("p_d", [(1, 1), (2, 3), (3, 4)])
def test_kmv_quad_tet_schemes(p_d):  # noqa: W503
    fct = math.factorial
    p, d = p_d
    q = create_quadrature(Te, p, "KMV")
    for i in range(d + 1):
        for j in range(d + 1 - i):
            for k in range(d + 1 - i - j):
                trueval = fct(i) * fct(j) * fct(k) / fct(i + j + k + 3)
                assert (
                    np.abs(
                        trueval -
                        q.integrate(lambda x: x[0] ** i * x[1] ** j * x[2] ** k)
                    ) <
                    1.0e-10
                )


@pytest.mark.parametrize("p_d", [(1, 1), (2, 3), (3, 5), (4, 7), (5, 9), (6, 11)])
def test_kmv_quad_tri_schemes(p_d):
    fct = math.factorial
    p, d = p_d
    q = create_quadrature(T, p, "KMV")
    for i in range(d + 1):
        for j in range(d + 1 - i):
            trueval = fct(i) * fct(j) / fct(i + j + 2)
            assert (
                np.abs(trueval - q.integrate(lambda x: x[0] ** i * x[1] ** j)) < 1.0e-10
            )


@pytest.mark.parametrize(
    "element_degree",
    [(KMV(T, 1), 1), (KMV(T, 2), 2), (KMV(T, 3), 3), (KMV(T, 4), 4), (KMV(T, 5), 5), (KMV(T, 6), 6)],
)
def test_Kronecker_property_tris(element_degree):
    """
    Evaluating the nodal basis at the special quadrature points should
    have a Kronecker property.  Also checks that the basis functions
    and quadrature points are given the same ordering.
    """
    element, degree = element_degree
    qr = create_quadrature(T, degree, scheme="KMV")
    (basis,) = element.tabulate(0, qr.get_points()).values()
    assert np.allclose(basis, np.eye(*basis.shape))


@pytest.mark.parametrize(
    "element_degree", [(KMV(Te, 1), 1), (KMV(Te, 2), 2), (KMV(Te, 3), 3)]
)
def test_Kronecker_property_tets(element_degree):
    """
    Evaluating the nodal basis at the special quadrature points should
    have a Kronecker property.  Also checks that the basis functions
    and quadrature points are given the same ordering.
    """
    element, degree = element_degree
    qr = create_quadrature(Te, degree, scheme="KMV")
    (basis,) = element.tabulate(0, qr.get_points()).values()
    assert np.allclose(basis, np.eye(*basis.shape))


@pytest.mark.parametrize("degree", [2, 3, 4])
def test_edge_degree(degree):
    """Verify that the outer edges of a degree KMV element
       are indeed of degree and the interior is of degree+1"""
    # create a degree+1 polynomial
    I = UFCInterval()
    # an exact quad. rule for a degree+1 polynomial on the UFCinterval
    qr = make_quadrature(I, degree + 1)
    W = np.diag(qr.wts)
    sd = I.get_spatial_dimension()
    pset = polynomial_set.ONPolynomialSet(I, degree + 1, (sd,))
    pset = pset.take([degree + 1])
    # tabulate at the quadrature points
    interval_vals = pset.tabulate(qr.get_points())[(0,)]
    interval_vals = np.squeeze(interval_vals)
    # create degree KMV element (should have degree outer edges and degree+1 edge in center)
    T = UFCTriangle()
    element = KMV(T, degree)
    # tabulate values on an edge of the KMV element
    for e in range(3):
        edge_values = element.tabulate(0, qr.get_points(), (1, e))[(0, 0)]
        # degree edge should be orthogonal to degree+1 ONpoly edge values
        result = edge_values @ W @ interval_vals.T
        assert np.allclose(np.sum(result), 0.0)


@pytest.mark.parametrize(
    "element_degree",
    [(KMV(T, 1), 1), (KMV(T, 2), 2), (KMV(T, 3), 3), (KMV(T, 4), 4), (KMV(T, 5), 5), (KMV(T, 6), 6)],
)
def test_interpolate_monomials_tris(element_degree):
    element, degree = element_degree

    # ordered the same way as KMV nodes
    pts = create_quadrature(T, degree, "KMV").pts

    Q = make_quadrature(T, 2 * degree)
    phis = element.tabulate(0, Q.pts)[0, 0]
    print("deg", degree)
    for i in range(degree + 1):
        for j in range(degree + 1 - i):
            m = lambda x: x[0] ** i * x[1] ** j
            dofs = np.array([m(pt) for pt in pts])
            interp = phis.T @ dofs
            matqp = np.array([m(pt) for pt in Q.pts])
            err = 0.0
            for k in range(phis.shape[1]):
                err += Q.wts[k] * (interp[k] - matqp[k]) ** 2
            assert np.sqrt(err) <= 1.0e-12


@pytest.mark.parametrize(
    "element_degree", [(KMV(Te, 1), 1), (KMV(Te, 2), 2), (KMV(Te, 3), 3)]
)
def test_interpolate_monomials_tets(element_degree):
    element, degree = element_degree

    # ordered the same way as KMV nodes
    pts = create_quadrature(Te, degree, "KMV").pts

    Q = make_quadrature(Te, 2 * degree)
    phis = element.tabulate(0, Q.pts)[0, 0, 0]
    print("deg", degree)
    for i in range(degree + 1):
        for j in range(degree + 1 - i):
            for k in range(degree + 1 - i - j):
                m = lambda x: x[0] ** i * x[1] ** j * x[2] ** k
                dofs = np.array([m(pt) for pt in pts])
                interp = phis.T @ dofs
                matqp = np.array([m(pt) for pt in Q.pts])
                err = 0.0
                for kk in range(phis.shape[1]):
                    err += Q.wts[kk] * (interp[kk] - matqp[kk]) ** 2
                assert np.sqrt(err) <= 1.0e-12



"""This module defines the UFL finite element classes."""

# Copyright (C) 2008-2016 Martin Sandve Alnæs
#
# This file was originally part of UFL (https://www.fenicsproject.org)
#
# SPDX-License-Identifier:    LGPL-3.0-or-later
#
# Modified by Kristian B. Oelgaard
# Modified by Marie E. Rognes 2010, 2012
# Modified by Massimiliano Leoni, 2016
# Modified by Matthew Scroggs, 2023

from finat.ufl.finiteelementbase import FiniteElementBase


class EnrichedElementBase(FiniteElementBase):
    """The vector sum of several finite element spaces."""

    def __init__(self, *elements):
        """Doc."""
        self._elements = elements

        cell = elements[0].cell
        if not all(e.cell == cell for e in elements[1:]):
            raise ValueError("Cell mismatch for sub elements of enriched element.")

        if isinstance(elements[0].degree(), int):
            degrees = {e.degree() for e in elements} - {None}
            degree = max(degrees) if degrees else None
        else:
            degree = tuple(map(max, zip(*[e.degree() for e in elements])))

        # We can allow the scheme not to be defined, but all defined
        # should be equal
        quad_schemes = [e.quadrature_scheme() for e in elements]
        quad_schemes = [qs for qs in quad_schemes if qs is not None]
        quad_scheme = quad_schemes[0] if quad_schemes else None
        if not all(qs == quad_scheme for qs in quad_schemes):
            raise ValueError("Quadrature scheme mismatch.")

        reference_value_shape = elements[0].reference_value_shape
        if not all(e.reference_value_shape == reference_value_shape for e in elements[1:]):
            raise ValueError("Element reference value shape mismatch.")

        # mapping = elements[0].mapping() # FIXME: This fails for a mixed subelement here.
        # if not all(e.mapping() == mapping for e in elements[1:]):
        #    raise ValueError("Element mapping mismatch.")

        # Get name of subclass: EnrichedElement or NodalEnrichedElement
        class_name = self.__class__.__name__

        # Initialize element data
        FiniteElementBase.__init__(self, class_name, cell, degree,
                                   quad_scheme, reference_value_shape)

    def mapping(self):
        """Doc."""
        return self._elements[0].mapping()

    @property
    def sobolev_space(self):
        """Return the underlying Sobolev space."""
        elements = [e for e in self._elements]
        if all(e.sobolev_space == elements[0].sobolev_space
               for e in elements):
            return elements[0].sobolev_space
        else:
            # Find smallest shared Sobolev space over all sub elements
            spaces = [e.sobolev_space for e in elements]
            superspaces = [{s} | set(s.parents) for s in spaces]
            intersect = set.intersection(*superspaces)
            for s in intersect.copy():
                for parent in s.parents:
                    intersect.discard(parent)

            sobolev_space, = intersect
            return sobolev_space

    def variant(self):
        """Doc."""
        try:
            variant, = {e.variant() for e in self._elements}
            return variant
        except ValueError:
            return None

    def reconstruct(self, **kwargs):
        """Doc."""
        return type(self)(*[e.reconstruct(**kwargs) for e in self._elements])

    @property
    def embedded_subdegree(self):
        """Return embedded subdegree."""
        return min(e.embedded_subdegree for e in self._elements)

    @property
    def embedded_superdegree(self):
        """Return embedded superdegree."""
        return max(e.embedded_superdegree for e in self._elements)


class EnrichedElement(EnrichedElementBase):
    r"""The vector sum of several finite element spaces.

    .. math:: \\textrm{EnrichedElement}(V, Q) = \\{v + q | v \\in V, q \\in Q\\}.

    Dual basis is a concatenation of subelements dual bases;
    primal basis is a concatenation of subelements primal bases;
    resulting element is not nodal even when subelements are.
    Structured basis may be exploited in form compilers.
    """

    def is_cellwise_constant(self):
        """Return whether the basis functions of this element is spatially constant over each cell."""
        return all(e.is_cellwise_constant() for e in self._elements)

    def __repr__(self):
        """Doc."""
        return "EnrichedElement(" + ", ".join(repr(e) for e in self._elements) + ")"

    def __str__(self):
        """Format as string for pretty printing."""
        return "<%s>" % " + ".join(str(e) for e in self._elements)

    def shortstr(self):
        """Format as string for pretty printing."""
        return "<%s>" % " + ".join(e.shortstr() for e in self._elements)


class NodalEnrichedElement(EnrichedElementBase):
    r"""The vector sum of several finite element spaces.

    .. math:: \\textrm{EnrichedElement}(V, Q) = \\{v + q | v \\in V, q \\in Q\\}.

    Primal basis is reorthogonalized to dual basis which is
    a concatenation of subelements dual bases; resulting
    element is nodal.
    """
    def is_cellwise_constant(self):
        """Return whether the basis functions of this element is spatially constant over each cell."""
        return False

    def __repr__(self):
        """Doc."""
        return "NodalEnrichedElement(" + ", ".join(repr(e) for e in self._elements) + ")"

    def __str__(self):
        """Format as string for pretty printing."""
        return "<Nodal enriched element(%s)>" % ", ".join(str(e) for e in self._elements)

    def shortstr(self):
        """Format as string for pretty printing."""
        return "NodalEnriched(%s)" % ", ".join(e.shortstr() for e in self._elements)



import pytest

import ufl
import finat.ufl
import finat
from finat.element_factory import create_element, supported_elements


@pytest.fixture(params=["BDM",
                        "BDFM",
                        "Lagrange",
                        "N1curl",
                        "N2curl",
                        "RT",
                        "Regge"])
def triangle_names(request):
    return request.param


@pytest.fixture
def ufl_element(triangle_names):
    return finat.ufl.FiniteElement(triangle_names, ufl.triangle, 2)


def test_triangle_basic(ufl_element):
    element = create_element(ufl_element)
    assert isinstance(element, supported_elements[ufl_element.family()])


@pytest.fixture
def ufl_vector_element(triangle_names):
    return finat.ufl.VectorElement(triangle_names, ufl.triangle, 2)


def test_triangle_vector(ufl_element, ufl_vector_element):
    scalar = create_element(ufl_element)
    vector = create_element(ufl_vector_element)

    assert isinstance(vector, finat.TensorFiniteElement)
    assert scalar == vector.base_element


@pytest.fixture(params=["CG", "DG", "DG L2"])
def tensor_name(request):
    return request.param


@pytest.fixture(params=[ufl.interval, ufl.triangle,
                        ufl.quadrilateral],
                ids=lambda x: x.cellname())
def ufl_A(request, tensor_name):
    return finat.ufl.FiniteElement(tensor_name, request.param, 1)


@pytest.fixture
def ufl_B(tensor_name):
    return finat.ufl.FiniteElement(tensor_name, ufl.interval, 1)


def test_tensor_prod_simple(ufl_A, ufl_B):
    tensor_ufl = finat.ufl.TensorProductElement(ufl_A, ufl_B)

    tensor = create_element(tensor_ufl)
    A = create_element(ufl_A)
    B = create_element(ufl_B)

    assert isinstance(tensor, finat.TensorProductElement)

    assert tensor.factors == (A, B)


@pytest.mark.parametrize(('family', 'expected_cls'),
                         [('P', finat.GaussLobattoLegendre),
                          ('DP', finat.GaussLegendre),
                          ('DP L2', finat.GaussLegendre)])
def test_interval_variant_default(family, expected_cls):
    ufl_element = finat.ufl.FiniteElement(family, ufl.interval, 3)
    assert isinstance(create_element(ufl_element), expected_cls)


@pytest.mark.parametrize(('family', 'variant', 'expected_cls'),
                         [('P', 'equispaced', finat.Lagrange),
                          ('P', 'spectral', finat.GaussLobattoLegendre),
                          ('DP', 'equispaced', finat.DiscontinuousLagrange),
                          ('DP', 'spectral', finat.GaussLegendre),
                          ('DP L2', 'equispaced', finat.DiscontinuousLagrange),
                          ('DP L2', 'spectral', finat.GaussLegendre)])
def test_interval_variant(family, variant, expected_cls):
    ufl_element = finat.ufl.FiniteElement(family, ufl.interval, 3, variant=variant)
    assert isinstance(create_element(ufl_element), expected_cls)


def test_triangle_variant_spectral():
    ufl_element = finat.ufl.FiniteElement('DP', ufl.triangle, 2, variant='spectral')
    create_element(ufl_element)


def test_triangle_variant_spectral_l2():
    ufl_element = finat.ufl.FiniteElement('DP L2', ufl.triangle, 2, variant='spectral')
    create_element(ufl_element)


def test_quadrilateral_variant_spectral_q():
    element = create_element(finat.ufl.FiniteElement('Q', ufl.quadrilateral, 3, variant='spectral'))
    assert isinstance(element.product.factors[0], finat.GaussLobattoLegendre)
    assert isinstance(element.product.factors[1], finat.GaussLobattoLegendre)


def test_quadrilateral_variant_spectral_dq():
    element = create_element(finat.ufl.FiniteElement('DQ', ufl.quadrilateral, 1, variant='spectral'))
    assert isinstance(element.product.factors[0], finat.GaussLegendre)
    assert isinstance(element.product.factors[1], finat.GaussLegendre)


def test_quadrilateral_variant_spectral_dq_l2():
    element = create_element(finat.ufl.FiniteElement('DQ L2', ufl.quadrilateral, 1, variant='spectral'))
    assert isinstance(element.product.factors[0], finat.GaussLegendre)
    assert isinstance(element.product.factors[1], finat.GaussLegendre)


def test_cache_hit(ufl_element):
    A = create_element(ufl_element)
    B = create_element(ufl_element)

    assert A is B


def test_cache_hit_vector(ufl_vector_element):
    A = create_element(ufl_vector_element)
    B = create_element(ufl_vector_element)

    assert A is B


if __name__ == "__main__":
    import os
    import sys
    pytest.main(args=[os.path.abspath(__file__)] + sys.argv[1:])



"""Data structures and algorithms for generic expansion and
refactorisation."""

from collections import Counter, OrderedDict, defaultdict, namedtuple
from functools import singledispatch
from itertools import product
from sys import intern

from gem.node import Memoizer, traversal
from gem.gem import (Node, Conditional, Zero, Product, Sum, Indexed,
                     ListTensor, one, MathFunction)
from gem.optimise import (remove_componenttensors, sum_factorise,
                          traverse_product, traverse_sum, unroll_indexsum,
                          make_rename_map, make_renamer)


# Refactorisation labels

ATOMIC = intern('atomic')
"""Label: the expression need not be broken up into smaller parts"""

COMPOUND = intern('compound')
"""Label: the expression must be broken up into smaller parts"""

OTHER = intern('other')
"""Label: the expression is irrelevant with regards to refactorisation"""


Monomial = namedtuple('Monomial', ['sum_indices', 'atomics', 'rest'])
"""Monomial type, representation of a tensor product with some
distinguished factors (called atomics).

- sum_indices: indices to sum over
- atomics: tuple of expressions classified as ATOMIC
- rest: a single expression classified as OTHER

A :py:class:`Monomial` is a structured description of the expression:

.. code-block:: python

    IndexSum(reduce(Product, atomics, rest), sum_indices)

"""


class MonomialSum(object):
    """Represents a sum of :py:class:`Monomial`s.

    The set of :py:class:`Monomial` summands are represented as a
    mapping from a pair of unordered ``sum_indices`` and unordered
    ``atomics`` to a ``rest`` GEM expression.  This representation
    makes it easier to merge similar monomials.
    """
    def __init__(self):
        # (unordered sum_indices, unordered atomics) -> rest
        self.monomials = defaultdict(Zero)

        # We shall retain ordering for deterministic code generation:
        #
        # (unordered sum_indices, unordered atomics) ->
        #     (ordered sum_indices, ordered atomics)
        self.ordering = OrderedDict()

    def __len__(self):
        return len(self.ordering)

    def add(self, sum_indices, atomics, rest):
        """Updates the :py:class:`MonomialSum` adding a new monomial."""
        sum_indices = tuple(sum_indices)
        sum_indices_set = frozenset(sum_indices)
        # Sum indices cannot have duplicates
        assert len(sum_indices) == len(sum_indices_set)

        atomics = tuple(atomics)
        atomics_set = frozenset(Counter(atomics).items())

        assert isinstance(rest, Node)

        key = (sum_indices_set, atomics_set)
        self.monomials[key] = Sum(self.monomials[key], rest)
        self.ordering.setdefault(key, (sum_indices, atomics))

    def __iter__(self):
        """Iteration yields :py:class:`Monomial` objects"""
        for key, (sum_indices, atomics) in self.ordering.items():
            rest = self.monomials[key]
            yield Monomial(sum_indices, atomics, rest)

    @staticmethod
    def sum(*args):
        """Sum of multiple :py:class:`MonomialSum`s"""
        result = MonomialSum()
        for arg in args:
            assert isinstance(arg, MonomialSum)
            # Optimised implementation: no need to decompose and
            # reconstruct key.
            for key, rest in arg.monomials.items():
                result.monomials[key] = Sum(result.monomials[key], rest)
            for key, value in arg.ordering.items():
                result.ordering.setdefault(key, value)
        return result

    @staticmethod
    def product(*args, **kwargs):
        """Product of multiple :py:class:`MonomialSum`s"""
        rename_map = kwargs.pop('rename_map', None)
        if rename_map is None:
            rename_map = make_rename_map()
        if kwargs:
            raise ValueError("Unrecognised keyword argument: " + kwargs.pop())

        result = MonomialSum()
        for monomials in product(*args):
            renamer = make_renamer(rename_map)
            sum_indices = []
            atomics = []
            rest = one
            for s, a, r in monomials:
                s_, applier = renamer(s)
                sum_indices.extend(s_)
                atomics.extend(map(applier, a))
                rest = Product(applier(r), rest)
            result.add(sum_indices, atomics, rest)
        return result


class FactorisationError(Exception):
    """Raised when factorisation fails to achieve some desired form."""
    pass


@singledispatch
def _collect_monomials(expression, self):
    """Refactorises an expression into a sum-of-products form, using
    distributivity rules (i.e. a*(b + c) -> a*b + a*c).  Expansion
    proceeds until all "compound" expressions are broken up.

    :arg expression: a GEM expression to refactorise
    :arg self: function for recursive calls

    :returns: :py:class:`MonomialSum`

    :raises FactorisationError: Failed to break up some "compound"
                                expressions with expansion.
    """
    # Phase 1: Collect and categorise product terms
    def stop_at(expr):
        # Break up compounds only
        return self.classifier(expr) != COMPOUND
    common_indices, terms = traverse_product(expression, stop_at=stop_at)
    common_indices = tuple(common_indices)

    common_atomics = []
    common_others = []
    compounds = []
    for term in terms:
        label = self.classifier(term)
        if label == ATOMIC:
            common_atomics.append(term)
        elif label == COMPOUND:
            compounds.append(term)
        elif label == OTHER:
            common_others.append(term)
        else:
            raise ValueError("Classifier returned illegal value.")
    common_atomics = tuple(common_atomics)

    # Phase 2: Attempt to break up compound terms into summands
    sums = []
    for expr in compounds:
        summands = traverse_sum(expr, stop_at=stop_at)
        if len(summands) <= 1 and not isinstance(expr, (Conditional, MathFunction)):
            # Compound term is not an addition, avoid infinite
            # recursion and fail gracefully raising an exception.
            raise FactorisationError(expr)
        # Recurse into each summand, concatenate their results
        sums.append(MonomialSum.sum(*map(self, summands)))

    # Phase 3: Expansion
    #
    # Each element of ``sums`` is a MonomialSum.  Expansion produces a
    # series (representing a sum) of products of monomials.
    result = MonomialSum()
    for s, a, r in MonomialSum.product(*sums, rename_map=self.rename_map):
        renamer = make_renamer(self.rename_map)
        renamer(common_indices)  # update current_set
        s_, applier = renamer(s)

        all_indices = common_indices + s_
        atomics = common_atomics + tuple(map(applier, a))

        # All free indices that appear in atomic terms
        atomic_indices = set().union(*[atomic.free_indices
                                       for atomic in atomics])

        # Sum indices that appear in atomic terms
        # (will go to the result :py:class:`Monomial`)
        sum_indices = tuple(index for index in all_indices
                            if index in atomic_indices)

        # Sum indices that do not appear in atomic terms
        # (can factorise them over atomic terms immediately)
        rest_indices = tuple(index for index in all_indices
                             if index not in atomic_indices)

        # Not really sum factorisation, but rather just an optimised
        # way of building a product.
        rest = sum_factorise(rest_indices, common_others + [applier(r)])

        result.add(sum_indices, atomics, rest)
    return result


@_collect_monomials.register(MathFunction)
def _collect_monomials_mathfunction(expression, self):
    name = expression.name
    if name in {"conj", "real", "imag"}:
        # These are allowed to be applied to arguments, and hence must
        # be dealt with specially. Just push the function onto each
        # entry in the monomialsum of the child.
        # NOTE: This presently assumes that the "atomics" part of a
        # MonomialSum are real. This is true for the coffee, tensor,
        # spectral modes: the atomics are indexed tabulation matrices
        # (which are guaranteed real).
        # If the classifier puts (potentially) complex expressions in
        # atomics, then this code needs fixed.
        child_ms, = map(self, expression.children)
        result = MonomialSum()
        for k, v in child_ms.monomials.items():
            result.monomials[k] = MathFunction(name, v)
        result.ordering = child_ms.ordering.copy()
        return result
    else:
        return _collect_monomials.dispatch(MathFunction.mro()[1])(expression, self)


@_collect_monomials.register(Conditional)
def _collect_monomials_conditional(expression, self):
    """Refactorises a conditional expression into a sum-of-products form,
    pulling only "atomics" out of conditional expressions.

    :arg expression: a GEM expression to refactorise
    :arg self: function for recursive calls

    :returns: :py:class:`MonomialSum`
    """
    condition, then, else_ = expression.children
    # Recursively refactorise both branches to `MonomialSum`s
    then_ms = self(then)
    else_ms = self(else_)

    result = MonomialSum()
    # For each set of atomics, create a new Conditional node.  Atomics
    # are considered safe to be pulled out of conditionals, but other
    # expressions remain inside conditional branches.
    zero = Zero()
    for k in then_ms.monomials.keys() | else_ms.monomials.keys():
        _then = then_ms.monomials.get(k, zero)
        _else = else_ms.monomials.get(k, zero)
        result.monomials[k] = Conditional(condition, _then, _else)

    # Construct a deterministic ordering
    result.ordering = then_ms.ordering.copy()
    for k, v in else_ms.ordering.items():
        result.ordering.setdefault(k, v)
    return result


def collect_monomials(expressions, classifier):
    """Refactorises expressions into a sum-of-products form, using
    distributivity rules (i.e. a*(b + c) -> a*b + a*c).  Expansion
    proceeds until all "compound" expressions are broken up.

    :arg expressions: GEM expressions to refactorise
    :arg classifier: a function that can classify any GEM expression
                     as ``ATOMIC``, ``COMPOUND``, or ``OTHER``.  This
                     classification drives the factorisation.

    :returns: list of :py:class:`MonomialSum`s

    :raises FactorisationError: Failed to break up some "compound"
                                expressions with expansion.
    """
    # Get ComponentTensors out of the way
    expressions = remove_componenttensors(expressions)

    # Get ListTensors out of the way
    must_unroll = []  # indices to unroll
    for node in traversal(expressions):
        if isinstance(node, Indexed):
            child, = node.children
            if isinstance(child, ListTensor) and classifier(node) == COMPOUND:
                must_unroll.extend(node.multiindex)
    if must_unroll:
        must_unroll = set(must_unroll)
        expressions = unroll_indexsum(expressions,
                                      predicate=lambda i: i in must_unroll)
        expressions = remove_componenttensors(expressions)

    # Finally, refactorise expressions
    mapper = Memoizer(_collect_monomials)
    mapper.classifier = classifier
    mapper.rename_map = make_rename_map()
    return list(map(mapper, expressions))



# This file is part of FIAT (https://www.fenicsproject.org)
#
# SPDX-License-Identifier:    LGPL-3.0-or-later
#
# Written by Pablo D. Brubeck (brubeck@protonmail.com), 2024

# This is not quite Christiansen-Hu, but it has 2*dim*(dim+1) dofs and includes
# dim**2-1 extra constraint functionals.  The first (dim+1)**2 basis functions
# are the reference element bfs, but the extra dim**2-1 are used in the
# transformation theory.

from FIAT import finite_element, polynomial_set
from FIAT.bernardi_raugel import BernardiRaugelDualSet
from FIAT.quadrature_schemes import create_quadrature
from FIAT.macro import CkPolynomialSet, WorseyFarinSplit

import numpy


def ChristiansenHuSpace(ref_el, degree, reduced=False):
    """Return a basis for the Christianse-Hu space
    set(v in C0 P1(WF)^d : div(v) = 0) + P_0 x if reduced = True, and
    this space is agumented with rotated facet bubbles if reduced = False."""
    sd = ref_el.get_spatial_dimension()
    ref_complex = WorseyFarinSplit(ref_el)
    C0 = CkPolynomialSet(ref_complex, degree, order=0, shape=(sd,), scale=1, variant="bubble")
    Q = create_quadrature(ref_complex, degree-1)
    tab = C0.tabulate(Q.get_points(), 1)
    divC0 = sum(tab[alpha][:, alpha.index(1), :] for alpha in tab if sum(alpha) == 1)

    nsp = polynomial_set.spanning_basis(divC0.T, nullspace=True)
    coeffs = numpy.tensordot(nsp, C0.get_coeffs(), axes=(-1, 0))

    verts = numpy.array(ref_complex.get_vertices())
    WT = verts[-1]
    P0x_coeffs = numpy.transpose(verts - WT[None, :])
    coeffs = numpy.concatenate((coeffs, P0x_coeffs[None, ...]), axis=0)

    if not reduced:
        # Compute the primal basis via Vandermonde and extract the facet bubbles
        dual = BernardiRaugelDualSet(ref_el, degree, degree=degree, ref_complex=ref_complex, reduced=True)
        dualmat = dual.to_riesz(C0)
        V = numpy.tensordot(dualmat, coeffs, axes=((1, 2), (1, 2)))
        coeffs = numpy.tensordot(numpy.linalg.inv(V.T), coeffs, axes=(-1, 0))
        facet_bubbles = coeffs[-(sd+1):]

        # Rotate the facet bubbles onto the tangent space of the facet
        # NOTE they are not aligned with the normal, but they point in the direction
        # that connects the split point on the facet with the split point of the cell
        WF = verts[sd+1:-1]
        top = ref_el.get_topology()
        ext = []
        for f in top[sd-1]:
            ehat = WF[f] - WT
            FB = numpy.dot(ehat, facet_bubbles[f])
            thats = ref_el.compute_tangents(sd-1, f)
            for that in thats:
                ext.append(that[:, None] * FB[None, :])
        ext_coeffs = numpy.array(ext)
        coeffs = numpy.concatenate((coeffs, ext_coeffs), axis=0)

    return polynomial_set.PolynomialSet(ref_complex, degree, degree,
                                        C0.get_expansion_set(), coeffs)


class ChristiansenHu(finite_element.CiarletElement):
    """The Christiansen-Hu C^0(Worsey-Farin) linear macroelement with divergence in P0.
    This element belongs to a Stokes complex, and is paired with unsplit DG0."""
    def __init__(self, ref_el, degree=1):
        if degree != 1:
            raise ValueError("Christiansen-Hu only defined for degree = 1")
        poly_set = ChristiansenHuSpace(ref_el, degree)
        ref_complex = poly_set.get_reference_element()
        dual = BernardiRaugelDualSet(ref_el, degree, degree=degree, ref_complex=ref_complex)
        formdegree = ref_el.get_spatial_dimension() - 1  # (n-1)-form
        super().__init__(poly_set, dual, degree, formdegree, mapping="contravariant piola")



# Copyright (C) 2005 The University of Chicago
#
# This file is part of FIAT (https://www.fenicsproject.org)
#
# SPDX-License-Identifier:    LGPL-3.0-or-later
#
# Written by Robert C. Kirby
# Modified by Andrew T. T. McRae (Imperial College London)
#
# This work is partially supported by the US Department of Energy
# under award number DE-FG02-04ER25650

from FIAT import dual_set, functional, polynomial_set, finite_element
import numpy


class P0Dual(dual_set.DualSet):
    def __init__(self, ref_el):
        entity_ids = {}
        entity_permutations = {}
        sd = ref_el.get_dimension()
        top = ref_el.get_topology()
        if sd == 0:
            pts = [tuple() for entity in sorted(top[sd])]
        else:
            pts = [tuple(numpy.average(ref_el.get_vertices_of_subcomplex(top[sd][entity]), 0))
                   for entity in sorted(top[sd])]
        nodes = [functional.PointEvaluation(ref_el, pt) for pt in pts]
        for dim in sorted(top):
            entity_ids[dim] = {}
            entity_permutations[dim] = {}
            sym_size = ref_el.symmetry_group_size(dim)
            num_points = 1 if dim == sd else 0
            if isinstance(dim, tuple):
                assert isinstance(sym_size, tuple)
                perms = {o: list(range(num_points)) for o in numpy.ndindex(sym_size)}
            else:
                perms = {o: list(range(num_points)) for o in range(sym_size)}
            for entity in sorted(top[dim]):
                entity_ids[dim][entity] = [entity] if dim == sd else []
                entity_permutations[dim][entity] = perms

        super().__init__(nodes, ref_el, entity_ids, entity_permutations)


class P0(finite_element.CiarletElement):
    def __init__(self, ref_el):
        poly_set = polynomial_set.ONPolynomialSet(ref_el, 0)
        dual = P0Dual(ref_el)
        degree = 0
        formdegree = ref_el.get_spatial_dimension()  # n-form
        super().__init__(poly_set, dual, degree, formdegree)



# Copyright (C) 2008-2012 Robert C. Kirby (Texas Tech University)
# Modified by Andrew T. T. McRae (Imperial College London)
#
# This file is part of FIAT (https://www.fenicsproject.org)
#
# SPDX-License-Identifier:    LGPL-3.0-or-later

from FIAT import (finite_element, functional, dual_set,
                  polynomial_set, nedelec)
from FIAT.check_format_variant import check_format_variant
from FIAT.quadrature_schemes import create_quadrature
from FIAT.quadrature import FacetQuadratureRule
import numpy


class BDMDualSet(dual_set.DualSet):
    def __init__(self, ref_el, degree, variant, interpolant_deg):
        nodes = []
        sd = ref_el.get_spatial_dimension()
        top = ref_el.get_topology()

        entity_ids = {}
        # set to empty
        for dim in top:
            entity_ids[dim] = {}
            for entity in top[dim]:
                entity_ids[dim][entity] = []

        if variant == "integral":
            facet = ref_el.construct_subelement(sd-1)
            # Facet nodes are \int_F v\cdot n p ds where p \in P_{q}
            # degree is q
            Q_ref = create_quadrature(facet, interpolant_deg + degree)
            Pq = polynomial_set.ONPolynomialSet(facet, degree)
            Pq_at_qpts = Pq.tabulate(Q_ref.get_points())[(0,)*(sd - 1)]
            for f in top[sd - 1]:
                cur = len(nodes)
                Q = FacetQuadratureRule(ref_el, sd - 1, f, Q_ref)
                Jdet = Q.jacobian_determinant()
                n = ref_el.compute_scaled_normal(f) / Jdet
                phis = n[None, :, None] * Pq_at_qpts[:, None, :]
                nodes.extend(functional.FrobeniusIntegralMoment(ref_el, Q, phi)
                             for phi in phis)
                entity_ids[sd - 1][f] = list(range(cur, len(nodes)))

        elif variant == "point":
            # Define each functional for the dual set
            # codimension 1 facets
            for f in top[sd - 1]:
                cur = len(nodes)
                pts_cur = ref_el.make_points(sd - 1, f, sd + degree)
                nodes.extend(functional.PointScaledNormalEvaluation(ref_el, f, pt)
                             for pt in pts_cur)
                entity_ids[sd - 1][f] = list(range(cur, len(nodes)))

        # internal nodes
        if degree > 1:
            if interpolant_deg is None:
                interpolant_deg = degree

            cell = ref_el.construct_subelement(sd)
            Q_ref = create_quadrature(cell, interpolant_deg + degree - 1)
            Nedel = nedelec.Nedelec(cell, degree - 1, variant)
            Ned_at_qpts = Nedel.tabulate(0, Q_ref.get_points())[(0,) * sd]
            for entity in top[sd]:
                Q = FacetQuadratureRule(ref_el, sd, entity, Q_ref)
                Jinv = numpy.linalg.inv(Q.jacobian())
                phis = numpy.tensordot(Jinv.T, Ned_at_qpts, (1, 1)).transpose((1, 0, 2))
                cur = len(nodes)
                nodes.extend(functional.FrobeniusIntegralMoment(ref_el, Q, phi) for phi in phis)
                entity_ids[sd][entity] = list(range(cur, len(nodes)))

        super().__init__(nodes, ref_el, entity_ids)


class BrezziDouglasMarini(finite_element.CiarletElement):
    """
    The BDM element

    :arg ref_el: The reference element.
    :arg degree: The degree.
    :arg variant: optional variant specifying the types of nodes.

    variant can be chosen from ["point", "integral", "integral(q)"]
    "point" -> dofs are evaluated by point evaluation. Note that this variant
    has suboptimal convergence order in the H(div)-norm
    "integral" -> dofs are evaluated by quadrature rules with the minimum
    degree required for unisolvence.
    "integral(q)" -> dofs are evaluated by quadrature rules with the minimum
    degree required for unisolvence plus q. You might want to choose a high
    quadrature degree to make sure that expressions will be interpolated
    exactly. This is important when you want to have (nearly) div-preserving
    interpolation.
    """

    def __init__(self, ref_el, degree, variant=None):

        splitting, variant, interpolant_deg = check_format_variant(variant, degree)
        if splitting is not None:
            ref_el = splitting(ref_el)

        if degree < 1:
            raise Exception("BDM_k elements only valid for k >= 1")

        sd = ref_el.get_spatial_dimension()
        poly_set = polynomial_set.ONPolynomialSet(ref_el, degree, (sd, ))
        dual = BDMDualSet(ref_el, degree, variant, interpolant_deg)
        formdegree = sd - 1  # (n-1)-form
        super().__init__(poly_set, dual, degree, formdegree, mapping="contravariant piola")



# Copyright (C) 2018 Robert C. Kirby
#
# This file is part of FIAT (https://www.fenicsproject.org)
#
# SPDX-License-Identifier:    LGPL-3.0-or-later

# This is not quite Bell, but is 21-dofs and includes 3 extra constraint
# functionals.  The first 18 basis functions are the reference element
# bfs, but the extra three are used in the transformation theory.

from FIAT import finite_element, polynomial_set, dual_set, functional
from FIAT.reference_element import TRIANGLE, ufc_simplex


class BellDualSet(dual_set.DualSet):
    def __init__(self, ref_el):
        entity_ids = {}
        nodes = []
        cur = 0

        # make nodes by getting points
        # need to do this dimension-by-dimension, facet-by-facet
        top = ref_el.get_topology()
        verts = ref_el.get_vertices()
        sd = ref_el.get_spatial_dimension()
        if ref_el.get_shape() != TRIANGLE:
            raise ValueError("Bell only defined on triangles")

        pd = functional.PointDerivative

        # get jet at each vertex

        entity_ids[0] = {}
        for v in sorted(top[0]):
            nodes.append(functional.PointEvaluation(ref_el, verts[v]))

            # first derivatives
            for i in range(sd):
                alpha = [0] * sd
                alpha[i] = 1
                nodes.append(pd(ref_el, verts[v], alpha))

            # second derivatives
            alphas = [[2, 0], [1, 1], [0, 2]]
            for alpha in alphas:
                nodes.append(pd(ref_el, verts[v], alpha))

            entity_ids[0][v] = list(range(cur, cur + 6))
            cur += 6

        # we need an edge quadrature rule for the moment
        from FIAT.quadrature_schemes import create_quadrature
        from FIAT.jacobi import eval_jacobi
        rline = ufc_simplex(1)
        q1d = create_quadrature(rline, 8)
        q1dpts = q1d.get_points()[:, 0]
        leg4_at_qpts = eval_jacobi(0, 0, 4, 2.0*q1dpts - 1)

        imond = functional.IntegralMomentOfNormalDerivative
        entity_ids[1] = {}
        for e in sorted(top[1]):
            entity_ids[1][e] = [18+e]
            nodes.append(imond(ref_el, e, q1d, leg4_at_qpts))

        entity_ids[2] = {0: []}

        super().__init__(nodes, ref_el, entity_ids)


class Bell(finite_element.CiarletElement):
    """The Bell finite element."""

    def __init__(self, ref_el):
        poly_set = polynomial_set.ONPolynomialSet(ref_el, 5)
        dual = BellDualSet(ref_el)
        super().__init__(poly_set, dual, 5)



import pytest
import numpy

from FIAT import RestrictedElement, HsiehCloughTocher as HCT
from FIAT.reference_element import ufc_simplex
from FIAT.functional import PointEvaluation
from FIAT.macro import CkPolynomialSet


@pytest.fixture
def cell():
    K = ufc_simplex(2)
    K.vertices = ((0.0, 0.1), (1.17, -0.09), (0.15, 1.84))
    return K


def span_greater_equal(A, B):
    # span(A) >= span(B)
    _, residual, *_ = numpy.linalg.lstsq(A.reshape(A.shape[0], -1).T,
                                         B.reshape(B.shape[0], -1).T)
    return numpy.allclose(residual, 0)


def make_points(K, degree):
    top = K.get_topology()
    pts = []
    for dim in top:
        for entity in top[dim]:
            pts.extend(K.make_points(dim, entity, degree))
    return pts


@pytest.mark.parametrize("reduced", (False, True))
def test_hct_constant(cell, reduced):
    # Test that bfs associated with point evaluation sum up to 1
    fe = HCT(cell, reduced=reduced)

    pts = make_points(cell, 4)
    tab = fe.tabulate(2, pts)

    coefs = numpy.zeros((fe.space_dimension(),))
    nodes = fe.dual_basis()
    entity_dofs = fe.entity_dofs()
    for v in entity_dofs[0]:
        for k in entity_dofs[0][v]:
            if isinstance(nodes[k], PointEvaluation):
                coefs[k] = 1.0

    for alpha in tab:
        expected = 1 if sum(alpha) == 0 else 0
        vals = numpy.dot(coefs, tab[alpha])
        assert numpy.allclose(vals, expected)


@pytest.mark.parametrize("reduced", (False, True))
def test_full_polynomials(cell, reduced):
    # Test that HCT/HCT-red contains all cubics/quadratics
    fe = HCT(cell, reduced=reduced)
    if reduced:
        fe = RestrictedElement(fe, restriction_domain="vertex")

    ref_complex = fe.get_reference_complex()
    pts = make_points(ref_complex, 4)
    tab = fe.tabulate(0, pts)[(0, 0)]

    degree = fe.degree()
    if reduced:
        degree -= 1

    P = CkPolynomialSet(cell, degree, variant="bubble")
    P_tab = P.tabulate(pts)[(0, 0)]
    assert span_greater_equal(tab, P_tab)

    C1 = CkPolynomialSet(ref_complex, degree, order=1, variant="bubble")
    C1_tab = C1.tabulate(pts)[(0, 0)]
    assert span_greater_equal(tab, C1_tab)



from __future__ import absolute_import, division, print_function

from FIAT.reference_element import (UFCHexahedron, UFCQuadrilateral,
                                    compute_unflattening_map, flatten_entities,
                                    flatten_permutations)
from FIAT.tensor_product import FlattenedDimensions as FIAT_FlattenedDimensions
from gem.utils import cached_property

from finat.finiteelementbase import FiniteElementBase


class FlattenedDimensions(FiniteElementBase):
    """Class for elements on quadrilaterals and hexahedra.  Wraps a tensor
    product element on a tensor product cell, and flattens its entity
    dimensions."""

    def __init__(self, element):
        super().__init__()
        self.product = element
        self._unflatten = compute_unflattening_map(element.cell.get_topology())

    @cached_property
    def cell(self):
        dim = self.product.cell.get_spatial_dimension()
        if dim == 2:
            return UFCQuadrilateral()
        elif dim == 3:
            return UFCHexahedron()
        else:
            raise NotImplementedError("Cannot guess cell for spatial dimension %s" % dim)

    @property
    def complex(self):
        return self.product.complex

    @property
    def degree(self):
        unique_degree, = set(self.product.degree)
        return unique_degree

    @property
    def formdegree(self):
        return self.product.formdegree

    @cached_property
    def _entity_dofs(self):
        return flatten_entities(self.product.entity_dofs())

    @cached_property
    def _entity_support_dofs(self):
        return flatten_entities(self.product.entity_support_dofs())

    def entity_dofs(self):
        return self._entity_dofs

    @cached_property
    def entity_permutations(self):
        return flatten_permutations(self.product.entity_permutations)

    def space_dimension(self):
        return self.product.space_dimension()

    @cached_property
    def fiat_equivalent(self):
        return FIAT_FlattenedDimensions(self.product.fiat_equivalent)

    def basis_evaluation(self, order, ps, entity=None, coordinate_mapping=None):
        """Return code for evaluating the element at known points on the
        reference element.

        :param order: return derivatives up to this order.
        :param ps: the point set object.
        :param entity: the cell entity on which to tabulate.
        """
        if entity is None:
            entity = (self.cell.get_spatial_dimension(), 0)

        return self.product.basis_evaluation(order, ps, self._unflatten[entity])

    def point_evaluation(self, order, point, entity=None, coordinate_mapping=None):
        if entity is None:
            entity = (self.cell.get_spatial_dimension(), 0)

        return self.product.point_evaluation(order, point, self._unflatten[entity],
                                             coordinate_mapping)

    @property
    def dual_basis(self):
        return self.product.dual_basis

    @property
    def index_shape(self):
        return self.product.index_shape

    @property
    def value_shape(self):
        return self.product.value_shape

    @property
    def mapping(self):
        return self.product.mapping



import math
import numpy
import pytest
from FIAT import (Lagrange, DiscontinuousLagrange, IntegratedLegendre, Legendre, P0,
                  Nedelec, NedelecSecondKind, RaviartThomas, BrezziDouglasMarini,
                  Regge, HellanHerrmannJohnson, GopalakrishnanLedererSchoberlSecondKind,
                  CrouzeixRaviart)
from FIAT.macro import AlfeldSplit, IsoSplit, PowellSabinSplit, CkPolynomialSet, MacroPolynomialSet
from FIAT.quadrature_schemes import create_quadrature
from FIAT.reference_element import ufc_simplex, UFCSimplex
from FIAT.expansions import polynomial_entity_ids, polynomial_cell_node_map
from FIAT.polynomial_set import make_bubbles, PolynomialSet, ONPolynomialSet
from FIAT.barycentric_interpolation import get_lagrange_points


@pytest.fixture(params=(1, 2, 3), ids=("I", "T", "S"))
def cell(request):
    dim = request.param
    return ufc_simplex(dim)


def test_split_cache(cell):
    A = AlfeldSplit(cell)
    B = AlfeldSplit(cell)
    assert B is A
    fe = Lagrange(cell, 1, variant="alfeld")
    C = fe.ref_complex
    assert C is A


@pytest.mark.parametrize("split", (AlfeldSplit, IsoSplit))
def test_split_entity_transform(split, cell):
    split_cell = split(cell)
    top = split_cell.get_topology()
    sdim = cell.get_spatial_dimension()
    for dim in range(1, sdim+1):
        ref_el = split_cell.construct_subelement(dim)
        b = numpy.average(ref_el.get_vertices(), axis=0)
        for entity in top[dim]:
            mapped_bary = split_cell.get_entity_transform(dim, entity)(b)
            computed_bary = numpy.average(split_cell.get_vertices_of_subcomplex(top[dim][entity]), axis=0)
            assert numpy.allclose(mapped_bary, computed_bary)


@pytest.mark.parametrize("degree", (4,))
@pytest.mark.parametrize("variant", ("gll", "equispaced"))
@pytest.mark.parametrize("split", (AlfeldSplit, IsoSplit))
def test_split_make_points(split, cell, degree, variant):
    split_cell = split(cell)
    top = split_cell.get_topology()
    sdim = cell.get_spatial_dimension()
    for i in range(1, sdim+1):
        ref_el = split_cell.construct_subelement(i)
        pts_ref = ref_el.make_points(i, 0, degree, variant=variant)
        for entity in top[i]:
            pts_entity = split_cell.make_points(i, entity, degree, variant=variant)
            mapping = split_cell.get_entity_transform(i, entity)
            mapped_pts = mapping(pts_ref)
            assert numpy.allclose(mapped_pts, pts_entity)


def test_split_child_to_parent(cell):
    split_cell = IsoSplit(cell)

    dim = cell.get_spatial_dimension()
    degree = 2 if dim == 3 else 4
    parent_degree = 2*degree

    top = cell.get_topology()
    parent_pts = {dim: {} for dim in top}
    for dim in top:
        for entity in top[dim]:
            parent_pts[dim][entity] = cell.make_points(dim, entity, parent_degree)

    top = split_cell.get_topology()
    child_pts = {dim: {} for dim in top}
    for dim in top:
        for entity in top[dim]:
            child_pts[dim][entity] = split_cell.make_points(dim, entity, degree)

    child_to_parent = split_cell.get_child_to_parent()
    for dim in top:
        for entity in top[dim]:
            parent_dim, parent_entity = child_to_parent[dim][entity]
            assert set(child_pts[dim][entity]) <= set(parent_pts[parent_dim][parent_entity])


@pytest.mark.parametrize("split", (AlfeldSplit, IsoSplit))
def test_macro_quadrature(split, cell):
    ref_el = split(cell)
    sd = ref_el.get_spatial_dimension()

    degree = 3
    Q = create_quadrature(ref_el, 2*degree)
    pts, wts = Q.get_points(), Q.get_weights()

    # Test that the mass matrix for an orthogonal basis is diagonal
    fe = Legendre(ref_el, degree)
    phis = fe.tabulate(0, pts)[(0,)*sd]
    M = numpy.dot(numpy.multiply(phis, wts), phis.T)
    M = M - numpy.diag(M.diagonal())
    assert numpy.allclose(M, 0)


@pytest.mark.parametrize("degree", range(1, 5))
@pytest.mark.parametrize("variant", ("equispaced", "gll"))
@pytest.mark.parametrize("split", (AlfeldSplit, IsoSplit))
def test_macro_lagrange(variant, degree, split, cell):
    ref_el = split(cell)

    fe = Lagrange(ref_el, degree, variant=variant)
    poly_set = fe.get_nodal_basis()

    # Test that the polynomial set is defined on the split and not on the parent cell
    assert poly_set.get_reference_element() is ref_el

    # Test that the finite element is defined on the parent cell and not on the split
    assert fe.get_reference_element() is cell

    # Test that parent entities are the ones exposed
    entity_ids = fe.entity_dofs()
    parent_top = ref_el.get_parent().get_topology()
    for dim in parent_top:
        assert len(entity_ids[dim]) == len(parent_top[dim])

    # Test that tabulation onto lattice points gives the identity
    sd = ref_el.get_spatial_dimension()
    parent_to_children = ref_el.get_parent_to_children()
    pts = []
    for dim in sorted(parent_to_children):
        for entity in sorted(parent_to_children[dim]):
            for cdim, centity in parent_to_children[dim][entity]:
                pts.extend(ref_el.make_points(cdim, centity, degree, variant=variant))

    phis = fe.tabulate(2, pts)
    assert numpy.allclose(phis[(0,)*sd], numpy.eye(fe.space_dimension()))

    # Test that we can reproduce the Vandermonde matrix by tabulating the expansion set
    U = poly_set.get_expansion_set()
    V = U.tabulate(degree, pts).T
    assert numpy.allclose(fe.V, V)


def test_powell_sabin(cell):
    dim = cell.get_spatial_dimension()
    A = AlfeldSplit(cell)
    assert A > cell

    PS = PowellSabinSplit(cell, dim)
    assert PS == A

    for split_dim in range(1, dim):
        PS = PowellSabinSplit(cell, split_dim)
        assert PS > A
        assert PS > cell
        assert len(PS.get_topology()[dim]) == math.factorial(dim+1) // math.factorial(split_dim)


def make_mass_matrix(fe, order=0):
    sd = fe.ref_el.get_spatial_dimension()
    Q = create_quadrature(fe.ref_complex, 2*fe.degree())
    qpts, qwts = Q.get_points(), Q.get_weights()
    phi = fe.tabulate(order, qpts)[(0,) * sd]
    M = numpy.dot(numpy.multiply(phi, qwts), phi.T)
    return M


@pytest.mark.parametrize("degree", (1, 2, 4))
@pytest.mark.parametrize("variant", ("equispaced", "gll"))
def test_lagrange_alfeld_duals(cell, degree, variant):
    Pk = Lagrange(cell, degree, variant=variant)
    alfeld = Lagrange(AlfeldSplit(cell), degree, variant=variant)

    Pk_pts = numpy.asarray(get_lagrange_points(Pk.dual_basis()))
    alfeld_pts = numpy.asarray(get_lagrange_points(alfeld.dual_basis()))
    ids = alfeld.entity_dofs()

    sd = cell.get_spatial_dimension()
    facet_dim = sum(len(ids[dim][entity]) for dim in range(sd) for entity in ids[dim])
    assert numpy.allclose(alfeld_pts[:facet_dim], Pk_pts[:facet_dim])

    phi = Pk.tabulate(0, alfeld_pts)[(0,) * sd]
    M_Pk = make_mass_matrix(Pk)
    M_alfeld = make_mass_matrix(alfeld)
    M_galerkin = numpy.dot(numpy.dot(phi, M_alfeld), phi.T)
    assert numpy.allclose(M_Pk, M_galerkin)


@pytest.mark.parametrize("degree", (1, 2, 4))
def test_lagrange_iso_duals(cell, degree):
    Pk = Lagrange(cell, 2*degree, variant="equispaced")
    Piso = Lagrange(IsoSplit(cell), degree, variant="equispaced")

    Pk_pts = numpy.asarray(get_lagrange_points(Pk.dual_basis()))
    Piso_pts = numpy.asarray(get_lagrange_points(Piso.dual_basis()))
    ids = Piso.entity_dofs()

    reorder = []
    for dim in ids:
        for entity in ids[dim]:
            reorder.extend(ids[dim][entity])
    assert numpy.allclose(Piso_pts[reorder], Pk_pts)

    poly_set = Piso.get_nodal_basis().take(reorder)
    assert numpy.allclose(numpy.eye(Piso.space_dimension()),
                          numpy.dot(Pk.get_dual_set().to_riesz(poly_set),
                                    poly_set.get_coeffs().T))


@pytest.mark.parametrize("variant", ("gll", "Alfeld,equispaced", "gll,iso"))
def test_is_macro_lagrange(variant):
    is_macro = "alfeld" in variant.lower() or "iso" in variant.lower()

    fe = Lagrange(ufc_simplex(2), 2, variant)
    assert not fe.get_reference_element().is_macrocell()
    assert fe.is_macroelement() == is_macro
    assert fe.get_reference_complex().is_macrocell() == is_macro
    assert fe.get_nodal_basis().get_reference_element().is_macrocell() == is_macro


@pytest.mark.parametrize("variant", ("gl", "Alfeld,equispaced_interior", "chebyshev,iso"))
@pytest.mark.parametrize("degree", (0, 2))
def test_is_macro_discontinuous_lagrange(degree, variant):
    is_macro = "alfeld" in variant.lower() or "iso" in variant.lower()

    fe = DiscontinuousLagrange(ufc_simplex(2), degree, variant)
    if degree == 0 and not is_macro:
        assert isinstance(fe, P0)
    assert not fe.get_reference_element().is_macrocell()
    assert fe.is_macroelement() == is_macro
    assert fe.get_reference_complex().is_macrocell() == is_macro
    assert fe.get_nodal_basis().get_reference_element().is_macrocell() == is_macro


@pytest.mark.parametrize('split', [None, AlfeldSplit])
@pytest.mark.parametrize('codim', range(3))
def test_make_bubbles(cell, split, codim):
    sd = cell.get_spatial_dimension()
    if codim > sd:
        return
    degree = 5
    if split is not None:
        cell = split(cell)
    B = make_bubbles(cell, degree, codim=codim)

    # basic tests
    assert isinstance(B, PolynomialSet)
    assert B.degree == degree
    num_members = B.get_num_members()
    top = cell.get_topology()
    assert num_members == math.comb(degree-1, sd-codim) * len(top[sd - codim])

    # tabulate onto a lattice
    points = []
    for dim in range(sd+1-codim):
        for entity in sorted(top[dim]):
            points.extend(cell.make_points(dim, entity, degree))
    values = B.tabulate(points)[(0,) * sd]

    # test that bubbles vanish on the boundary
    num_pts_on_facet = len(points) - num_members
    facet_values = values[:, :num_pts_on_facet]
    assert numpy.allclose(facet_values, 0, atol=1E-12)

    # test linear independence
    interior_values = values[:, num_pts_on_facet:]
    assert numpy.linalg.matrix_rank(interior_values.T, tol=1E-12) == num_members

    # test block diagonal tabulation
    bubbles_per_entity = num_members // len(top[sd-codim])
    for entity in top[sd-codim]:
        i0 = entity * bubbles_per_entity
        i1 = (entity+1) * bubbles_per_entity
        assert numpy.allclose(interior_values[i0:i1, :i0], 0, atol=1E-12)
        assert numpy.allclose(interior_values[i0:i1, i1:], 0, atol=1E-12)

    # test trace similarity
    dim = sd - codim
    nfacets = len(top[dim])
    if nfacets > 1 and dim > 0:
        ref_facet = cell.construct_subelement(dim)
        ref_bubbles = make_bubbles(ref_facet, degree)
        ref_points = ref_facet.make_points(dim, 0, degree)
        ref_values = ref_bubbles.tabulate(ref_points)[(0,) * dim]

        scale = None
        bubbles_per_entity = ref_bubbles.get_num_members()
        cur = 0
        for entity in sorted(top[dim]):
            indices = list(range(cur, cur + bubbles_per_entity))
            cur_values = interior_values[numpy.ix_(indices, indices)]
            if scale is None:
                scale = numpy.max(abs(cur_values)) / numpy.max(abs(ref_values))
            assert numpy.allclose(ref_values * scale, cur_values)
            cur += bubbles_per_entity

    # test that bubbles do not have components in span(P_{degree+2} \ P_{degree})
    Pkdim = math.comb(degree + sd, sd)
    entity_ids = polynomial_entity_ids(cell, degree + 2)
    indices = []
    for entity in top[sd]:
        indices.extend(entity_ids[sd][entity][Pkdim:])
    P = ONPolynomialSet(cell, degree + 2)
    P = P.take(indices)

    Q = create_quadrature(cell, P.degree + B.degree)
    qpts, qwts = Q.get_points(), Q.get_weights()
    P_at_qpts = P.tabulate(qpts)[(0,) * sd]
    B_at_qpts = B.tabulate(qpts)[(0,) * sd]
    assert numpy.allclose(numpy.dot(numpy.multiply(P_at_qpts, qwts), B_at_qpts.T), 0.0)


@pytest.mark.parametrize("degree", (4,))
@pytest.mark.parametrize("variant", (None, "bubble"))
@pytest.mark.parametrize("split", (AlfeldSplit, IsoSplit))
def test_macro_expansion(cell, split, variant, degree):
    ref_complex = split(cell)
    top = ref_complex.get_topology()
    sd = ref_complex.get_spatial_dimension()
    P = ONPolynomialSet(ref_complex, degree, variant=variant, scale=1)

    npoints = degree + sd + 1
    cell_point_map = []
    pts = []
    for cell in top[sd]:
        cur = len(pts)
        pts.extend(ref_complex.make_points(sd, cell, npoints))
        cell_point_map.append(list(range(cur, len(pts))))

    order = 2
    values = P.tabulate(pts, order)
    cell_node_map = polynomial_cell_node_map(ref_complex, degree, continuity=P.expansion_set.continuity)
    for cell in top[sd]:
        sub_el = ref_complex.construct_subelement(sd)
        sub_el.vertices = ref_complex.get_vertices_of_subcomplex(top[sd][cell])
        Pcell = ONPolynomialSet(sub_el, degree, variant=variant, scale=1)

        cell_pts = sub_el.make_points(sd, 0, npoints)
        cell_values = Pcell.tabulate(cell_pts, order)

        ibfs = cell_node_map[cell]
        ipts = cell_point_map[cell]
        indices = numpy.ix_(ibfs, ipts)
        for alpha in values:
            assert numpy.allclose(cell_values[alpha], values[alpha][indices])


@pytest.mark.parametrize("order", (0, 1))
@pytest.mark.parametrize("variant", (None, "bubble"))
@pytest.mark.parametrize("degree", (1, 4))
def test_Ck_basis(cell, order, degree, variant):
    # Test that we can correctly tabulate on points on facets.
    # This breaks if we were binning points into more than one cell without a partition of unity.
    # It suffices to tabulate on the vertices of the simplicial complex.
    A = AlfeldSplit(cell)
    Ck = CkPolynomialSet(A, degree, order=order, variant=variant)
    U = Ck.get_expansion_set()
    cell_node_map = U.get_cell_node_map(degree)

    sd = A.get_spatial_dimension()
    top = A.get_topology()
    coeffs = Ck.get_coeffs()
    phis = Ck.tabulate(A.get_vertices())[(0,)*sd]

    for cell in top[sd]:
        ipts = list(top[sd][cell])
        verts = A.get_vertices_of_subcomplex(top[sd][cell])
        Uvals = U._tabulate_on_cell(degree, verts, 0, cell=cell)[(0,)*sd]
        local_phis = numpy.dot(coeffs[:, cell_node_map[cell]], Uvals)
        assert numpy.allclose(local_phis, phis[:, ipts])


def test_C2_double_alfeld():
    # Construct the quintic C2 spline on the double Alfeld split
    # See Section 7.5 of Lai & Schumacher
    K = ufc_simplex(2)
    DCT = AlfeldSplit(AlfeldSplit(K))

    degree = 5

    # C3 on major split facets, C2 elsewhere
    order = {}
    order[1] = dict.fromkeys(DCT.get_interior_facets(1), 2)
    order[1].update(dict.fromkeys(range(3, 6), 3))

    # C4 at minor split barycenters, C3 at major split barycenter
    order[0] = dict.fromkeys(DCT.get_interior_facets(0), 4)
    order[0][3] = 3

    P = CkPolynomialSet(DCT, degree, order=order, variant="bubble")
    assert P.get_num_members() == 27


def test_distance_to_point_l1(cell):
    A = AlfeldSplit(cell)
    dim = A.get_spatial_dimension()
    top = A.get_topology()
    p0, = cell.make_points(dim, 0, dim+1)

    # construct one point in front of each facet
    pts = []
    expected = []
    parent_top = cell.get_topology()
    for i in parent_top[dim-1]:
        Fi, = numpy.asarray(cell.make_points(dim-1, i, dim))
        n = cell.compute_normal(i)
        n *= numpy.dot(n, Fi - p0)
        n /= numpy.linalg.norm(n)
        d = 0.222 + i/10
        pts.append(Fi + d * n)
        expected.append(d)

    # the computed L1 distance agrees with the L2 distance for points in front of facets
    parent_distance = cell.distance_to_point_l1(pts, rescale=True)
    assert numpy.allclose(parent_distance, expected)

    # assert that the subcell measures the same distance as the parent
    for i in top[dim]:
        subcell_distance = A.distance_to_point_l1(pts, entity=(dim, i), rescale=True)
        assert numpy.isclose(subcell_distance[i], expected[i])
        assert all(subcell_distance[:i] > expected[:i])
        assert all(subcell_distance[i+1:] > expected[i+1:])


@pytest.mark.parametrize("element", (DiscontinuousLagrange, Lagrange))
def test_macro_sympy(cell, element):
    import sympy
    variant = "spectral,alfeld"
    K = IsoSplit(cell)
    ebig = element(K, 3, variant=variant)
    pts = get_lagrange_points(ebig.dual_basis())

    dim = cell.get_spatial_dimension()
    X = tuple(sympy.Symbol(f"X[{i}]") for i in range(dim))
    degrees = range(1, 3) if element is Lagrange else range(3)
    for degree in degrees:
        fe = element(cell, degree, variant=variant)
        tab_sympy = fe.tabulate(0, X)[(0,) * dim]

        phis = sympy.lambdify(X, tab_sympy)
        results = phis(*numpy.transpose(pts))
        tab_numpy = fe.tabulate(0, pts)[(0,) * dim]
        assert numpy.allclose(results, tab_numpy)


@pytest.mark.parametrize("element,degree", [
    (Lagrange, 1), (Nedelec, 1), (RaviartThomas, 1), (DiscontinuousLagrange, 0),
    (Regge, 0), (HellanHerrmannJohnson, 0),
    (GopalakrishnanLedererSchoberlSecondKind, 0)])
@pytest.mark.parametrize("dim", (2, 3))
def test_macro_polynomial_set(dim, element, degree):
    K = ufc_simplex(dim)
    A = IsoSplit(K)

    fe = element(K, degree)
    mapping = fe.mapping()[0]
    fdim = fe.formdegree
    if isinstance(fdim, tuple):
        fdim = max(fdim)
    comps = []
    pts = []
    for entity in A.topology[fdim]:
        pts_cur = A.make_points(fdim, entity, 1+fdim)
        pts.extend(pts_cur)

        if mapping == "affine":
            comp = numpy.ones(())
        elif mapping == "covariant contravariant piola":
            ts = A.compute_tangents(fdim, entity)
            n = A.compute_scaled_normal(entity)
            comp = ts[..., None] * n[None, None, :]

        elif mapping.endswith("covariant piola"):
            comp = A.compute_edge_tangent(entity)
        elif mapping.endswith("contravariant piola"):
            comp = A.compute_scaled_normal(entity)
        if mapping.startswith("double"):
            comp = numpy.outer(comp, comp)

        comps.extend(comp for pt in pts_cur)

    P = MacroPolynomialSet(A, fe)
    phis = P.tabulate(pts)[(0,)*dim]
    shape = phis.shape
    shp = shape[1:-1]
    ncomp = comp.shape[:-len(shp)]

    result = numpy.zeros((shape[0], *ncomp, shape[-1]))
    ax = (tuple(range(-len(shp), 0)), )*2
    for i, comp in enumerate(comps):
        result[..., i] = numpy.tensordot(comp, phis[..., i], ax).T
    result[abs(result) < 1E-14] = 0
    result = result[:len(pts)*numpy.prod(ncomp, dtype=int)]

    if ncomp:
        result = result.transpose((*range(1, len(ncomp)+1), 0, -1))
        result = result.reshape((shape[0], -1))
    assert numpy.allclose(numpy.diag(numpy.diag(result)), result)


def compare_macro_variant(element, K, degree, variant):
    """Test CiarletElement by comparing macro variant to their non-macro
       counterparts."""
    # Test unisolvence
    fe_macro = element(K, degree, variant=variant)

    # Ensure that we have a macroelement
    ref_complex = fe_macro.get_reference_complex()
    assert ref_complex.is_macrocell()
    top = ref_complex.topology
    dim = ref_complex.get_spatial_dimension()
    assert len(top[dim]) > 1

    # Construct the non-macro element on a subcell
    subcell = max(top[dim])
    vids = list(top[dim][subcell])
    subtop = {d: dict(enumerate(tuple(map(vids.index, top[d][e]))
                                for e in top[d]
                                if set(top[d][e]) <= set(vids)))
              for d in sorted(top)}
    vertices = ref_complex.get_vertices_of_subcomplex(vids)
    Ksub = UFCSimplex(K.shape, vertices, subtop)
    fe_ref = element(Ksub, degree)

    # Compute Vandermonde matrix on the subcell
    P = fe_macro.get_nodal_basis()
    B = P.get_coeffs()
    A = fe_ref.dual.to_riesz(P)
    V = numpy.tensordot(A, B, axes=(range(1, A.ndim), range(1, B.ndim)))

    # Assert that V = permutation matrix
    V[abs(V) < 1E-12] = 0
    rows, cols = numpy.nonzero(V)
    assert len(rows) == fe_ref.space_dimension()
    assert numpy.allclose(V[rows, cols], 1)

    # Test tabulation on interior points
    pts = Ksub.make_points(dim, 0, degree+dim, interior=1)
    tab_macro = fe_macro.tabulate(1, pts)
    tab_ref = fe_ref.tabulate(1, pts)
    for alpha in tab_ref:
        expected = tab_ref[alpha][rows]
        result = tab_macro[alpha][cols]
        assert numpy.allclose(result, expected)


@pytest.mark.parametrize("element,degree", [
    (IntegratedLegendre, 4), (Legendre, 1),
    (Nedelec, 3), (NedelecSecondKind, 3),
    (RaviartThomas, 2), (BrezziDouglasMarini, 2),
    (Regge, 3), (HellanHerrmannJohnson, 1),
    (GopalakrishnanLedererSchoberlSecondKind, 1),
])
@pytest.mark.parametrize("dim", (2, 3))
@pytest.mark.parametrize("variant", ("alfeld", "iso"))
def test_macro_variants(dim, element, degree, variant):
    K = ufc_simplex(dim)
    compare_macro_variant(element, K, degree, variant)


@pytest.mark.parametrize("element,degree", [
    (CrouzeixRaviart, 3),
])
@pytest.mark.parametrize("variant", ("alfeld", "iso"))
def test_macro_variants_triangle(element, degree, variant):
    K = ufc_simplex(2)
    compare_macro_variant(element, K, degree, variant)



# -*- coding: utf-8 -*-
"""Implementation of the Hellan-Herrmann-Johnson finite elements."""

# Copyright (C) 2016-2018 Lizao Li <lzlarryli@gmail.com>
#
# Modified by Pablo D. Brubeck (brubeck@protonmail.com), 2024
#
# This file is part of FIAT (https://www.fenicsproject.org)
#
# SPDX-License-Identifier:    LGPL-3.0-or-later
from FIAT import dual_set, finite_element, polynomial_set
from FIAT.check_format_variant import check_format_variant
from FIAT.functional import (PointwiseInnerProductEvaluation,
                             ComponentPointEvaluation,
                             TensorBidirectionalIntegralMoment as BidirectionalMoment)
from FIAT.quadrature import FacetQuadratureRule
from FIAT.quadrature_schemes import create_quadrature


class HellanHerrmannJohnsonDual(dual_set.DualSet):
    def __init__(self, ref_el, degree, variant, qdegree):
        sd = ref_el.get_spatial_dimension()
        top = ref_el.get_topology()
        entity_ids = {dim: {i: [] for i in sorted(top[dim])} for dim in sorted(top)}
        nodes = []

        cell_to_faces = ref_el.get_connectivity()[(sd, sd-1)]
        n = list(map(ref_el.compute_scaled_normal, sorted(top[sd-1])))
        if variant == "point":
            for f in sorted(top[sd-1]):
                cur = len(nodes)
                # n^T u n evaluated on a Pk lattice
                pts = ref_el.make_points(sd-1, f, degree + sd)
                nodes.extend(PointwiseInnerProductEvaluation(ref_el, n[f], n[f], pt)
                             for pt in pts)
                entity_ids[sd-1][f].extend(range(cur, len(nodes)))

            if sd == 2:
                # FIXME Keeping Cartesian dofs in 2D just to make regression test pass
                for entity in sorted(top[sd]):
                    faces = cell_to_faces[entity]
                    cur = len(nodes)
                    pts = ref_el.make_points(sd, entity, degree + sd)
                    nodes.extend(ComponentPointEvaluation(ref_el, (i, j), (sd, sd), pt)
                                 for i in range(sd) for j in range(i, sd) for pt in pts)
                    entity_ids[sd][entity].extend(range(cur, len(nodes)))
            else:
                for entity in sorted(top[sd]):
                    faces = cell_to_faces[entity]
                    cur = len(nodes)
                    # n[f]^T u n[f] evaluated on a P_{k-1} lattice
                    pts = ref_el.make_points(sd, entity, degree + sd)
                    nodes.extend(PointwiseInnerProductEvaluation(ref_el, n[f], n[f], pt)
                                 for pt in pts for f in faces)

                    # n[i+1]^T u n[i+2] evaluated on a Pk lattice
                    pts = ref_el.make_points(sd, entity, degree + sd + 1)
                    nodes.extend(PointwiseInnerProductEvaluation(ref_el, n[faces[i+1]], n[faces[i+2]], pt)
                                 for pt in pts for i in range((sd-1)*(sd-2)))
                    entity_ids[sd][entity].extend(range(cur, len(nodes)))

        elif variant == "integral":
            # Face dofs
            ref_facet = ref_el.construct_subelement(sd-1)
            Q_ref = create_quadrature(ref_facet, qdegree + degree)
            P = polynomial_set.ONPolynomialSet(ref_facet, degree)
            Phis = P.tabulate(Q_ref.get_points())[(0,)*(sd-1)]

            for f in sorted(top[sd-1]):
                cur = len(nodes)
                Q = FacetQuadratureRule(ref_el, sd-1, f, Q_ref)
                detJ = Q.jacobian_determinant()
                # n[f]^T u n[f] integrated against a basis for Pk
                nodes.extend(BidirectionalMoment(ref_el, n[f], n[f]/detJ, Q, phi) for phi in Phis)
                entity_ids[sd-1][f].extend(range(cur, len(nodes)))

            ref_facet = ref_el.construct_subelement(sd)
            Q_ref = create_quadrature(ref_facet, qdegree + degree)
            P = polynomial_set.ONPolynomialSet(ref_facet, degree)
            Phis = P.tabulate(Q_ref.get_points())[(0,) * sd]
            dimPkm1 = P.expansion_set.get_num_members(degree-1)

            # Interior dofs
            for entity in sorted(top[sd]):
                cur = len(nodes)
                faces = cell_to_faces[entity]
                Q = FacetQuadratureRule(ref_el, sd, entity, Q_ref)
                detJ = Q.jacobian_determinant()
                # n[f]^T u n[f] integrated against a basis for P_{k-1}
                nodes.extend(BidirectionalMoment(ref_el, n[f], n[f]/detJ, Q, phi)
                             for phi in Phis[:dimPkm1] for f in faces)
                # n[i+1]^T u n[i+2] integrated against a basis for Pk
                nodes.extend(BidirectionalMoment(ref_el, n[faces[i+1]], n[faces[i+2]]/detJ, Q, phi)
                             for phi in Phis for i in range((sd-1)*(sd-2)))
                entity_ids[sd][entity].extend(range(cur, len(nodes)))

        super().__init__(nodes, ref_el, entity_ids)


class HellanHerrmannJohnson(finite_element.CiarletElement):
    """The definition of Hellan-Herrmann-Johnson element.
       HHJ(k) is the space of symmetric-matrix-valued polynomials of degree k
       or less with normal-normal continuity.
    """
    def __init__(self, ref_el, degree=0, variant=None):
        if degree < 0:
            raise ValueError(f"{type(self).__name__} only defined for degree >= 0")

        splitting, variant, qdegree = check_format_variant(variant, degree)
        if splitting is not None:
            ref_el = splitting(ref_el)

        poly_set = polynomial_set.ONSymTensorPolynomialSet(ref_el, degree)
        dual = HellanHerrmannJohnsonDual(ref_el, degree, variant, qdegree)
        sd = ref_el.get_spatial_dimension()
        formdegree = (sd-1, sd-1)
        mapping = "double contravariant piola"
        super().__init__(poly_set, dual, degree, formdegree, mapping=mapping)



# Copyright (C) 2015 Imperial College London and others.
#
# This file is part of FIAT (https://www.fenicsproject.org)
#
# SPDX-License-Identifier:    LGPL-3.0-or-later
#
# Written by Pablo D. Brubeck (brubeck@protonmail.com), 2021

import abc
import numpy

from FIAT import dual_set, finite_element, functional, quadrature
from FIAT.barycentric_interpolation import LagrangePolynomialSet
from FIAT.hierarchical import IntegratedLegendre
from FIAT.orientation_utils import make_entity_permutations_simplex
from FIAT.P0 import P0Dual
from FIAT.reference_element import LINE


def sym_eig(A, B):
    """
    A numpy-only implementation of `scipy.linalg.eigh`
    """
    Linv = numpy.linalg.inv(numpy.linalg.cholesky(B))
    C = numpy.dot(Linv, numpy.dot(A, Linv.T))
    Z, V = numpy.linalg.eigh(C, "U")
    V = numpy.dot(Linv.T, V)
    return Z, V


def tridiag_eig(A, B):
    """
    Same as sym_eig, but assumes that A is already diagonal and B tri-diagonal
    """
    a = numpy.reciprocal(A.diagonal())
    numpy.sqrt(a, out=a)
    C = numpy.multiply(a, B)
    numpy.multiply(C, a[:, None], out=C)
    Z, V = numpy.linalg.eigh(C, "U")
    numpy.reciprocal(Z, out=Z)
    numpy.multiply(numpy.sqrt(Z), V, out=V)
    numpy.multiply(V, a[:, None], out=V)
    return Z, V


class FDMDual(dual_set.DualSet):
    """The dual basis for 1D elements with FDM shape functions."""
    def __init__(self, ref_el, degree, bc_order=1, formdegree=0, orthogonalize=False):
        # Define the generalized eigenproblem on a reference element
        embedded_degree = degree + formdegree
        embedded = IntegratedLegendre(ref_el, embedded_degree)
        edim = embedded.space_dimension()
        self.embedded = embedded

        vertices = ref_el.get_vertices()
        if bc_order == 1 and formdegree == 0:
            rule = quadrature.GaussLobattoLegendreQuadratureLineRule(ref_el, edim+1)
        else:
            rule = quadrature.GaussLegendreQuadratureLineRule(ref_el, edim)
        self.rule = rule

        solve_eig = sym_eig
        if bc_order == 1:
            solve_eig = tridiag_eig

        # Tabulate the BC nodes
        if bc_order == 0:
            C = numpy.empty((0, edim), "d")
        else:
            constraints = embedded.tabulate(bc_order-1, vertices)
            C = numpy.transpose(numpy.column_stack(list(constraints.values())))
        bdof = slice(None, C.shape[0])
        idof = slice(C.shape[0], None)

        # Tabulate the basis that splits the DOFs into interior and bcs
        E = numpy.eye(edim)
        E[bdof, idof] = -C[:, idof]
        E[bdof, :] = numpy.linalg.solve(C[:, bdof], E[bdof, :])

        # Assemble the constrained Galerkin matrices on the reference cell
        k = max(1, bc_order)
        phi = embedded.tabulate(k, rule.get_points())
        wts = rule.get_weights()
        E0 = numpy.dot(E.T, phi[(0, )])
        Ek = numpy.dot(E.T, phi[(k, )])
        B = numpy.dot(numpy.multiply(E0, wts), E0.T)
        A = numpy.dot(numpy.multiply(Ek, wts), Ek.T)

        # Eigenfunctions in the constrained basis
        S = numpy.eye(A.shape[0])
        lam = numpy.ones((A.shape[0],))
        if S.shape[0] > C.shape[0]:
            lam[idof], Sii = solve_eig(A[idof, idof], B[idof, idof])
            S[idof, idof] = Sii
            S[idof, bdof] = numpy.dot(Sii, numpy.dot(Sii.T, -B[idof, bdof]))

        if orthogonalize:
            Abb = numpy.dot(S[:, bdof].T, numpy.dot(A, S[:, bdof]))
            Bbb = numpy.dot(S[:, bdof].T, numpy.dot(B, S[:, bdof]))
            _, Qbb = sym_eig(Abb, Bbb)
            S[:, bdof] = numpy.dot(S[:, bdof], Qbb)

        if formdegree == 0:
            # Tabulate eigenbasis
            basis = numpy.dot(S.T, E0)
        else:
            # Tabulate the derivative of the eigenbasis and normalize
            if bc_order == 0:
                idof = lam > 1.0E-12
                lam[~idof] = 1.0E0
            numpy.reciprocal(lam, out=lam)
            numpy.sqrt(lam, out=lam)
            numpy.multiply(S, lam, out=S)
            basis = numpy.dot(S.T, Ek)

        bc_nodes = []
        if formdegree == 0:
            if orthogonalize:
                idof = slice(None)
            elif bc_order > 0:
                bc_nodes += [functional.PointEvaluation(ref_el, x) for x in vertices]
                for alpha in range(1, bc_order):
                    bc_nodes += [functional.PointDerivative(ref_el, x, [alpha]) for x in vertices]

        elif bc_order > 0:
            basis[bdof, :] = numpy.sqrt(1.0E0 / ref_el.volume())
            idof = slice(formdegree, None)

        nodes = bc_nodes + [functional.IntegralMoment(ref_el, rule, f) for f in basis[idof]]

        if len(bc_nodes) > 0:
            entity_ids = {0: {0: [0], 1: [1]},
                          1: {0: list(range(2, degree+1))}}
            entity_permutations = {}
            entity_permutations[0] = {0: {0: [0]}, 1: {0: [0]}}
            entity_permutations[1] = {0: make_entity_permutations_simplex(1, degree - 1)}
        else:
            entity_ids = {0: {0: [], 1: []},
                          1: {0: list(range(0, degree+1))}}
            entity_permutations = {}
            entity_permutations[0] = {0: {0: []}, 1: {0: []}}
            entity_permutations[1] = {0: make_entity_permutations_simplex(1, degree + 1)}
        super().__init__(nodes, ref_el, entity_ids, entity_permutations)


class FDMFiniteElement(finite_element.CiarletElement):
    """1D element that diagonalizes bilinear forms with BCs."""

    _orthogonalize = False

    @property
    @abc.abstractmethod
    def _bc_order(self):
        pass

    @property
    @abc.abstractmethod
    def _formdegree(self):
        pass

    def __init__(self, ref_el, degree):
        if ref_el.shape != LINE:
            raise ValueError("%s is only defined in one dimension." % type(self))
        if degree == 0:
            dual = P0Dual(ref_el)
        else:
            dual = FDMDual(ref_el, degree, bc_order=self._bc_order,
                           formdegree=self._formdegree, orthogonalize=self._orthogonalize)
        if self._formdegree == 0:
            poly_set = dual.embedded.poly_set
        else:
            lr = quadrature.GaussLegendreQuadratureLineRule(ref_el, degree+1)
            poly_set = LagrangePolynomialSet(ref_el, lr.get_points())
        super().__init__(poly_set, dual, degree, self._formdegree)


class FDMLagrange(FDMFiniteElement):
    """1D CG element with interior shape functions that diagonalize the Laplacian."""
    _bc_order = 1
    _formdegree = 0


class FDMDiscontinuousLagrange(FDMFiniteElement):
    """1D DG element with derivatives of interior CG FDM shape functions."""
    _bc_order = 1
    _formdegree = 1


class FDMQuadrature(FDMFiniteElement):
    """1D DG element with interior CG FDM shape functions and orthogonalized vertex modes."""
    _bc_order = 1
    _formdegree = 0
    _orthogonalize = True


class FDMBrokenH1(FDMFiniteElement):
    """1D DG element with shape functions that diagonalize the Laplacian."""
    _bc_order = 0
    _formdegree = 0


class FDMBrokenL2(FDMFiniteElement):
    """1D DG element with the derivates of DG FDM shape functions."""
    _bc_order = 0
    _formdegree = 1


class FDMHermite(FDMFiniteElement):
    """1D CG element with interior shape functions that diagonalize the biharmonic operator."""
    _bc_order = 2
    _formdegree = 0



import numpy
from math import comb

import FIAT

from gem import Literal, ListTensor

from finat.fiat_elements import ScalarFiatElement
from finat.physically_mapped import Citations, identity, PhysicallyMappedElement


def _vertex_transform(V, vorder, fiat_cell, coordinate_mapping):
    """Basis transformation for evaluation, gradient, and hessian at vertices."""
    sd = fiat_cell.get_spatial_dimension()
    top = fiat_cell.get_topology()
    bary, = fiat_cell.make_points(sd, 0, sd+1)
    J = coordinate_mapping.jacobian_at(bary)

    gdofs = sd
    G = [[J[j, i] for j in range(sd)] for i in range(sd)]

    if vorder < 2:
        hdofs = 0
        H = [[]]
    else:
        hdofs = (sd*(sd+1))//2
        indices = [(i, j) for i in range(sd) for j in range(i, sd)]
        H = numpy.zeros((hdofs, hdofs), dtype=object)
        for p, (i, j) in enumerate(indices):
            for q, (m, n) in enumerate(indices):
                H[p, q] = J[m, i] * J[n, j] + J[m, j] * J[n, i]
        H[:, [i == j for i, j in indices]] *= 0.5

    s = 0
    for v in sorted(top[0]):
        s += 1
        V[s:s+gdofs, s:s+gdofs] = G
        s += gdofs
        V[s:s+hdofs, s:s+hdofs] = H
        s += hdofs
    return V


def _normal_tangential_transform(fiat_cell, J, detJ, f):
    R = numpy.array([[0, 1], [-1, 0]])
    that = fiat_cell.compute_edge_tangent(f)
    nhat = R @ that
    Jn = J @ Literal(nhat)
    Jt = J @ Literal(that)
    alpha = Jn @ Jt
    beta = Jt @ Jt
    Bnn = detJ / beta
    Bnt = alpha / beta

    Lhat = numpy.linalg.norm(that)
    Bnn = Bnn * Lhat
    Bnt = Bnt / Lhat
    return Bnn, Bnt, Jt


def _edge_transform(V, vorder, eorder, fiat_cell, coordinate_mapping, avg=False):
    """Basis transformation for integral edge moments.

    :arg V: the transpose of the basis transformation.
    :arg vorder: the jet order at vertices, matching the Jacobi weights in the
                 normal derivative moments on edges.
    :arg eorder: the order of the normal derivative moments.
    :arg fiat_cell: the reference triangle.
    :arg coordinate_mapping: the coordinate mapping.
    :kwarg avg: are we scaling integrals by dividing by the edge length?
    """
    sd = fiat_cell.get_spatial_dimension()
    bary, = fiat_cell.make_points(sd, 0, sd+1)
    J = coordinate_mapping.jacobian_at(bary)
    detJ = coordinate_mapping.detJ_at(bary)
    pel = coordinate_mapping.physical_edge_lengths()

    # number of DOFs per vertex/edge
    voffset = comb(sd + vorder, vorder)
    eoffset = 2 * eorder + 1
    top = fiat_cell.get_topology()
    for e in sorted(top[1]):
        Bnn, Bnt, Jt = _normal_tangential_transform(fiat_cell, J, detJ, e)
        if avg:
            Bnn = Bnn * pel[e]

        v0id, v1id = (v * voffset for v in top[1][e])
        s0 = len(top[0]) * voffset + e * eoffset
        for k in range(eorder+1):
            s = s0 + k
            # Jacobi polynomial at the endpoints
            P1 = comb(k + vorder, k)
            P0 = -(-1)**k * P1
            V[s, s] = Bnn
            V[s, v1id] = P1 * Bnt
            V[s, v0id] = P0 * Bnt
            if k > 0:
                V[s, s + eorder] = -1 * Bnt


class Argyris(PhysicallyMappedElement, ScalarFiatElement):
    def __init__(self, cell, degree=5, variant=None, avg=False):
        if Citations is not None:
            Citations().register("Argyris1968")
        if variant is None:
            variant = "integral"
        if variant == "point" and degree != 5:
            raise NotImplementedError("Degree must be 5 for 'point' variant of Argyris")
        fiat_element = FIAT.Argyris(cell, degree, variant=variant)
        self.variant = variant
        self.avg = avg
        super().__init__(fiat_element)

    def basis_transformation(self, coordinate_mapping):
        sd = self.cell.get_spatial_dimension()
        top = self.cell.get_topology()

        V = identity(self.space_dimension())

        vorder = 2
        voffset = comb(sd + vorder, vorder)
        eorder = self.degree - 5

        _vertex_transform(V, vorder, self.cell, coordinate_mapping)
        if self.variant == "integral":
            _edge_transform(V, vorder, eorder, self.cell, coordinate_mapping, avg=self.avg)
        else:
            bary, = self.cell.make_points(sd, 0, sd+1)
            J = coordinate_mapping.jacobian_at(bary)
            detJ = coordinate_mapping.detJ_at(bary)
            pel = coordinate_mapping.physical_edge_lengths()
            for e in sorted(top[1]):
                s = len(top[0]) * voffset + e * (eorder+1)
                v0id, v1id = (v * voffset for v in top[1][e])
                Bnn, Bnt, Jt = _normal_tangential_transform(self.cell, J, detJ, e)

                # edge midpoint normal derivative
                V[s, s] = Bnn * pel[e]

                # vertex points
                V[s, v1id] = 15/8 * Bnt
                V[s, v0id] = -1 * V[s, v1id]

                # vertex derivatives
                for i in range(sd):
                    V[s, v1id+1+i] = -7/16 * Bnt * Jt[i]
                    V[s, v0id+1+i] = V[s, v1id+1+i]

                # second derivatives
                tau = [Jt[0]*Jt[0], 2*Jt[0]*Jt[1], Jt[1]*Jt[1]]
                for i in range(len(tau)):
                    V[s, v1id+3+i] = 1/32 * Bnt * tau[i]
                    V[s, v0id+3+i] = -1 * V[s, v1id+3+i]

        # Patch up conditioning
        h = coordinate_mapping.cell_size()
        for v in sorted(top[0]):
            s = voffset*v + 1
            V[:, s:s+sd] *= 1 / h[v]
            V[:, s+sd:voffset*(v+1)] *= 1 / (h[v]*h[v])

        if self.variant == "point":
            eoffset = 2 * eorder + 1
            for e in sorted(top[1]):
                v0, v1 = top[1][e]
                s = len(top[0]) * voffset + e * eoffset
                V[:, s:s+eorder+1] *= 2 / (h[v0] + h[v1])

        return ListTensor(V.T)



# Copyright (C) 2008-2012 Robert C. Kirby (Texas Tech University)
# Modified by Andrew T. T. McRae (Imperial College London)
#
# This file is part of FIAT (https://www.fenicsproject.org)
#
# SPDX-License-Identifier:    LGPL-3.0-or-later

from FIAT import (expansions, polynomial_set, dual_set,
                  finite_element, functional, macro)
import numpy
from itertools import chain
from FIAT.check_format_variant import check_format_variant
from FIAT.quadrature_schemes import create_quadrature
from FIAT.quadrature import FacetQuadratureRule


def RTSpace(ref_el, degree):
    """Constructs a basis for the Raviart-Thomas space
    (P_{degree-1})^d + P_{degree-1} x"""
    sd = ref_el.get_spatial_dimension()

    k = degree - 1
    vec_Pkp1 = polynomial_set.ONPolynomialSet(ref_el, k + 1, (sd,))

    dimPkp1 = expansions.polynomial_dimension(ref_el, k + 1)
    dimPk = expansions.polynomial_dimension(ref_el, k)
    dimPkm1 = expansions.polynomial_dimension(ref_el, k - 1)

    vec_Pk_indices = list(chain(*(range(i * dimPkp1, i * dimPkp1 + dimPk)
                                  for i in range(sd))))
    vec_Pk_from_Pkp1 = vec_Pkp1.take(vec_Pk_indices)

    Pkp1 = polynomial_set.ONPolynomialSet(ref_el, k + 1)
    PkH = Pkp1.take(list(range(dimPkm1, dimPk)))

    Q = create_quadrature(ref_el, 2 * (k + 1))
    Qpts, Qwts = Q.get_points(), Q.get_weights()

    # have to work on this through "tabulate" interface
    # first, tabulate PkH at quadrature points
    PkH_at_Qpts = PkH.tabulate(Qpts)[(0,) * sd]

    Pkp1_at_Qpts = Pkp1.tabulate(Qpts)[(0,) * sd]

    x = Qpts.T
    PkHx_at_Qpts = PkH_at_Qpts[:, None, :] * x[None, :, :]
    PkHx_coeffs = numpy.dot(numpy.multiply(PkHx_at_Qpts, Qwts), Pkp1_at_Qpts.T)
    PkHx = polynomial_set.PolynomialSet(ref_el,
                                        k,
                                        k + 1,
                                        vec_Pkp1.get_expansion_set(),
                                        PkHx_coeffs)
    return polynomial_set.polynomial_set_union_normalized(vec_Pk_from_Pkp1, PkHx)


class RTDualSet(dual_set.DualSet):
    """Dual basis for Raviart-Thomas elements consisting of point
    evaluation of normals on facets of codimension 1 and internal
    moments against polynomials"""

    def __init__(self, ref_el, degree, variant, interpolant_deg):
        nodes = []
        sd = ref_el.get_spatial_dimension()
        top = ref_el.get_topology()

        entity_ids = {}
        # set to empty
        for dim in top:
            entity_ids[dim] = {}
            for entity in top[dim]:
                entity_ids[dim][entity] = []

        if variant == "integral":
            facet = ref_el.construct_subelement(sd-1)
            # Facet nodes are \int_F v\cdot n p ds where p \in P_q
            q = degree - 1
            Q_ref = create_quadrature(facet, interpolant_deg + q)
            Pq = polynomial_set.ONPolynomialSet(facet, q if sd > 1 else 0)
            Pq_at_qpts = Pq.tabulate(Q_ref.get_points())[(0,)*(sd - 1)]
            for f in top[sd - 1]:
                cur = len(nodes)
                Q = FacetQuadratureRule(ref_el, sd-1, f, Q_ref)
                Jdet = Q.jacobian_determinant()
                n = ref_el.compute_scaled_normal(f) / Jdet
                phis = n[None, :, None] * Pq_at_qpts[:, None, :]
                nodes.extend(functional.FrobeniusIntegralMoment(ref_el, Q, phi)
                             for phi in phis)
                entity_ids[sd - 1][f] = list(range(cur, len(nodes)))

            # internal nodes. These are \int_T v \cdot p dx where p \in P_{q-1}^d
            if q > 0:
                cell = ref_el.construct_subelement(sd)
                Q_ref = create_quadrature(cell, interpolant_deg + q - 1)
                Pqm1 = polynomial_set.ONPolynomialSet(cell, q - 1)
                Pqm1_at_qpts = Pqm1.tabulate(Q_ref.get_points())[(0,) * sd]

                for entity in top[sd]:
                    Q = FacetQuadratureRule(ref_el, sd, entity, Q_ref)
                    cur = len(nodes)
                    nodes.extend(functional.IntegralMoment(ref_el, Q, phi, (d,), (sd,))
                                 for d in range(sd)
                                 for phi in Pqm1_at_qpts)
                    entity_ids[sd][entity] = list(range(cur, len(nodes)))

        elif variant == "point":
            # codimension 1 facets
            for i in top[sd - 1]:
                cur = len(nodes)
                pts_cur = ref_el.make_points(sd - 1, i, sd + degree - 1)
                nodes.extend(functional.PointScaledNormalEvaluation(ref_el, i, pt)
                             for pt in pts_cur)
                entity_ids[sd - 1][i] = list(range(cur, len(nodes)))

            # internal nodes.  Let's just use points at a lattice
            if degree > 1:
                cur = len(nodes)
                pts = ref_el.make_points(sd, 0, sd + degree - 1)
                nodes.extend(functional.ComponentPointEvaluation(ref_el, d, (sd,), pt)
                             for d in range(sd)
                             for pt in pts)
                entity_ids[sd][0] = list(range(cur, len(nodes)))

        super().__init__(nodes, ref_el, entity_ids)


class RaviartThomas(finite_element.CiarletElement):
    """
    The Raviart Thomas element

    :arg ref_el: The reference element.
    :arg degree: The degree.
    :arg variant: optional variant specifying the types of nodes.

    variant can be chosen from ["point", "integral", "integral(q)"]
    "point" -> dofs are evaluated by point evaluation. Note that this variant
    has suboptimal convergence order in the H(div)-norm
    "integral" -> dofs are evaluated by quadrature rules with the minimum
    degree required for unisolvence.
    "integral(q)" -> dofs are evaluated by quadrature rules with the minimum
    degree required for unisolvence plus q. You might want to choose a high
    quadrature degree to make sure that expressions will be interpolated
    exactly. This is important when you want to have (nearly) div-preserving
    interpolation.
    """

    def __init__(self, ref_el, degree, variant=None):
        splitting, variant, interpolant_deg = check_format_variant(variant, degree)
        if splitting is not None:
            ref_el = splitting(ref_el)

        if ref_el.is_macrocell():
            base_element = RaviartThomas(ref_el.get_parent(), degree)
            poly_set = macro.MacroPolynomialSet(ref_el, base_element)
        else:
            poly_set = RTSpace(ref_el, degree)
        dual = RTDualSet(ref_el, degree, variant, interpolant_deg)
        formdegree = ref_el.get_spatial_dimension() - 1  # (n-1)-form
        super().__init__(poly_set, dual, degree, formdegree, mapping="contravariant piola")



from abc import ABCMeta, abstractmethod

import gem
import numpy

try:
    from firedrake_citations import Citations
    Citations().add("Kirby2018zany", """
@Article{Kirby2018zany,
  author =       {Robert C. Kirby},
  title =        {A general approach to transforming finite elements},
  journal =      {SMAI Journal of Computational Mathematics},
  year =         2018,
  volume =       4,
  pages =        {197-224},
  doi =          {10.5802/smai-jcm.33},
  archiveprefix ={arXiv},
  eprint =       {1706.09017},
  primaryclass = {math.NA}
}
""")
    Citations().add("Kirby2019zany", """
@Article{Kirby:2019,
  author =       {Robert C. Kirby and Lawrence Mitchell},
  title =        {Code generation for generally mapped finite
                  elements},
  journal =      {ACM Transactions on Mathematical Software},
  year =         2019,
  volume =       45,
  number =       41,
  pages =        {41:1--41:23},
  doi =          {10.1145/3361745},
  archiveprefix ={arXiv},
  eprint =       {1808.05513},
  primaryclass = {cs.MS}
}""")
    Citations().add("Clough1965", """
@inproceedings{Clough1965,
  author =       {R. W. Clough, J. L. Tocher},
  title =        {Finite element stiffness matrices for analysis of plate bending},
  booktitle =    {Proc. of the First Conf. on Matrix Methods in Struct. Mech},
  year =         1965,
  pages =        {515-546},
}
""")
    Citations().add("Argyris1968", """
@Article{Argyris1968,
  author =       {J. H. Argyris and I. Fried and D. W. Scharpf},
  title =        {{The TUBA family of plate elements for the matrix
                  displacement method}},
  journal =      {The Aeronautical Journal},
  year =         1968,
  volume =       72,
  pages =        {701-709},
  doi =          {10.1017/S000192400008489X}
}
""")
    Citations().add("Bell1969", """
@Article{Bell1969,
  author =       {Kolbein Bell},
  title =        {A refined triangular plate bending finite element},
  journal =      {International Journal for Numerical Methods in
                  Engineering},
  year =         1969,
  volume =       1,
  number =       1,
  pages =        {101-122},
  doi =          {10.1002/nme.1620010108}
}
""")
    Citations().add("Ciarlet1972", r"""
@Article{Ciarlet1972,
  author =       {P. G. Ciarlet and P. A. Raviart},
  title =        {{General Lagrange and Hermite interpolation in
                  $\mathbb{R}^n$ with applications to finite element
                  methods}},
  journal =      {Archive for Rational Mechanics and Analysis},
  year =         1972,
  volume =       46,
  number =       3,
  pages =        {177-199},
  doi =          {10.1007/BF0025245}
}
""")
    Citations().add("Morley1971", """
@Article{Morley1971,
  author =       {L. S. D. Morley},
  title =        {The constant-moment plate-bending element},
  journal =      {The Journal of Strain Analysis for Engineering
                  Design},
  year =         1971,
  volume =       6,
  number =       1,
  pages =        {20-24},
  doi =          {10.1243/03093247V061020}
}
""")
    Citations().add("MingXu2006", """
@article{MingXu2006,
  doi = {10.1007/s00211-005-0662-x},
  title={{The Morley element for fourth order elliptic equations in any dimensions}},
  author={Ming, Wang and Xu, Jinchao},
  journal={Numerische Mathematik},
  volume={103},
  number={1},
  pages={155--169},
  year={2006},
  publisher={Springer}
}
""")
    Citations().add("Mardal2002", """
@article{Mardal2002,
        doi = {10.1137/s0036142901383910},
        year = 2002,
        volume = {40},
        number = {5},
        pages = {1605--1631},
        author = {Mardal, K.-A.~ and Tai, X.-C.~ and Winther, R.~},
        title = {A robust finite element method for {Darcy--Stokes} flow},
        journal = {{SIAM} Journal on Numerical Analysis}
}
""")
    Citations().add("Arnold2002", """
@article{Arnold2002,
        doi = {10.1007/s002110100348},
        year = 2002,
        volume = {92},
        number = {3},
        pages = {401--419},
        author = {Arnold, R.~N.~ and Winther, R.~},
        title = {Mixed finite elements for elasticity},
        journal = {Numerische Mathematik}
}
""")
    Citations().add("Arnold2003", """
@article{arnold2003,
        doi = {10.1142/s0218202503002507},
        year = 2003,
        volume = {13},
        number = {03},
        pages = {295--307},
        author = {Arnold, D.~N.~ and Winther, R.~},
        title = {Nonconforming mixed elements for elasticity},
        journal = {Mathematical Models and Methods in Applied Sciences}
}
""")
    Citations().add("Hu2015", """
@article{Hu2015,
      author = {Hu, J.~ and Zhang, S.~},
      title = {A family of conforming mixed finite elements for linear elasticity on triangular grids},
      year = {2015},
      month = jan,
      archiveprefix = {arXiv},
      eprint = {1406.7457},
}
""")
    Citations().add("Arbogast2017", """
@techreport{Arbogast2017,
  title={Direct serendipity finite elements on convex quadrilaterals},
  author={Arbogast, T and Tao, Z},
  year={2017},
  institution={Tech. Rep. ICES REPORT 17-28, Institute for Computational Engineering and Sciences}
}
""")
    Citations().add("Gopalakrishnan2024", """
@article{gopalakrishnan2024johnson,
  title={{The Johnson-Mercier elasticity element in any dimensions}},
  author={Gopalakrishnan, J and Guzman, J and Lee, J J},
  journal={arXiv preprint arXiv:2403.13189},
  year={2024}
}
""")
    Citations().add("Groselj2022", """
@article{groselj2022generalized,
  title={{Generalized C1 Clough--Tocher splines for CAGD and FEM}},
  author={Gro{\v{s}}elj, Jan and Knez, Marjeta},
  journal={Computer Methods in Applied Mechanics and Engineering},
  volume={395},
  pages={114983},
  year={2022},
  publisher={Elsevier}
}
""")
    Citations().add("PowellSabin1977", """
@article{powell1977piecewise,
  title={Piecewise quadratic approximations on triangles},
  author={Powell, Michael JD and Sabin, Malcolm A},
  journal={ACM Transactions on Mathematical Software},
  volume={3},
  number={4},
  pages={316--325},
  year={1977},
  publisher={ACM New York, NY, USA}
}
""")
    Citations().add("AlfeldSorokina2016", """
@article{alfeld2016linear,
  title={Linear differential operators on bivariate spline spaces and spline vector fields},
  author={Alfeld, Peter and Sorokina, Tatyana},
  journal={BIT Numerical Mathematics},
  volume={56},
  number={1},
  pages={15--32},
  year={2016},
  publisher={Springer}
}
""")
    Citations().add("ArnoldQin1992", """
@article{arnold1992quadratic,
  title={{Quadratic velocity/linear pressure Stokes elements}},
  author={Arnold, Douglas N and Qin, Jinshui},
  journal={Advances in computer methods for partial differential equations},
  volume={7},
  pages={28--34},
  year={1992}
}
""")
    Citations().add("ChristiansenHu2019", """
@article{christiansen2019finite,
  title={A finite element for Stokes with a commuting diagram },
  author={Christiansen, Snorre H and Hu, Kaibo},
  journal={Mathematical Analysis in Fluid and Gas Dynamics},
  volume={2107},
  pages={172--183},
  year={2019}
}
""")
    Citations().add("GuzmanNeilan2018", """
@article{guzman2018infsup,
    author = {Guzm\'{a}n, Johnny and Neilan, Michael},
    title = {{Inf-Sup Stable Finite Elements on Barycentric Refinements Producing Divergence--Free Approximations in Arbitrary Dimensions}},
    journal = {SIAM Journal on Numerical Analysis},
    volume = {56},
    number = {5},
    pages = {2826-2844},
    year = {2018},
    doi = {10.1137/17M1153467}
}
""")
    Citations().add("BernardiRaugel1985", """
@article{bernardi-raugel-0,
    AUTHOR = {Bernardi, Christine and Raugel, Genevi\\`eve},
     TITLE = {Analysis of some finite elements for the {Stokes} problem},
   JOURNAL = {Mathematics of Computation},
    VOLUME = {44},
      YEAR = {1985},
       DOI = {10.1090/S0025-5718-1985-0771031-7},
     PAGES = {{71--79}}
}
""")


except ImportError:
    Citations = None


class NeedsCoordinateMappingElement(metaclass=ABCMeta):
    """Abstract class for elements that require physical information
    either to map or construct their basis functions."""
    pass


class PhysicallyMappedElement(NeedsCoordinateMappingElement):
    """A mixin that applies a "physical" transformation to tabulated
    basis functions."""

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        if Citations is not None:
            Citations().register("Kirby2018zany")
            Citations().register("Kirby2019zany")

    @abstractmethod
    def basis_transformation(self, coordinate_mapping):
        """Transformation matrix for the basis functions.

        :arg coordinate_mapping: Object providing physical geometry."""
        pass

    def map_tabulation(self, tabulation, coordinate_mapping):
        assert coordinate_mapping is not None

        M = self.basis_transformation(coordinate_mapping)
        # we expect M to be sparse with O(1) nonzeros per row
        # for each row, get the column index of each nonzero entry
        csr = [[j for j in range(M.shape[1]) if not isinstance(M.array[i, j], gem.Zero)]
               for i in range(M.shape[0])]

        def matvec(table):
            # basis recombination using hand-rolled sparse-dense matrix multiplication
            table = [gem.partial_indexed(table, (j,)) for j in range(M.shape[1])]
            # the sum approach is faster than calling numpy.dot or gem.IndexSum
            expressions = [sum(M.array[i, j] * table[j] for j in js) for i, js in enumerate(csr)]
            val = gem.ListTensor(expressions)
            return gem.optimise.aggressive_unroll(val)

        return {alpha: matvec(tabulation[alpha]) for alpha in tabulation}

    def basis_evaluation(self, order, ps, entity=None, coordinate_mapping=None):
        result = super().basis_evaluation(order, ps, entity=entity)
        return self.map_tabulation(result, coordinate_mapping)

    def point_evaluation(self, order, refcoords, entity=None, coordinate_mapping=None):
        result = super().point_evaluation(order, refcoords, entity=entity)
        return self.map_tabulation(result, coordinate_mapping)


class DirectlyDefinedElement(NeedsCoordinateMappingElement):
    """Base class for directly defined elements such as direct
    serendipity that bypass a coordinate mapping."""
    pass


class PhysicalGeometry(metaclass=ABCMeta):

    @abstractmethod
    def cell_size(self):
        """The cell size at each vertex.

        :returns: A GEM expression for the cell size, shape (nvertex, ).
        """

    @abstractmethod
    def jacobian_at(self, point):
        """The jacobian of the physical coordinates at a point.

        :arg point: The point in reference space (on the cell) to
             evaluate the Jacobian.
        :returns: A GEM expression for the Jacobian, shape (gdim, tdim).
        """

    @abstractmethod
    def detJ_at(self, point):
        """The determinant of the jacobian of the physical coordinates at a point.

        :arg point: The point in reference space to evaluate the Jacobian determinant.
        :returns: A GEM expression for the Jacobian determinant.
        """

    @abstractmethod
    def reference_normals(self):
        """The (unit) reference cell normals for each facet.

        :returns: A GEM expression for the normal to each
           facet (numbered according to FIAT conventions), shape
           (nfacet, tdim).
        """

    @abstractmethod
    def physical_normals(self):
        """The (unit) physical cell normals for each facet.

        :returns: A GEM expression for the normal to each
           facet (numbered according to FIAT conventions).  These are
           all computed by a clockwise rotation of the physical
           tangents, shape (nfacet, gdim).
        """

    @abstractmethod
    def physical_tangents(self):
        """The (unit) physical cell tangents on each facet.

        :returns: A GEM expression for the tangent to each
           facet (numbered according to FIAT conventions).  These
           always point from low to high numbered local vertex, shape
           (nfacet, gdim).
        """

    @abstractmethod
    def physical_edge_lengths(self):
        """The length of each edge of the physical cell.

        :returns: A GEM expression for the length of each
           edge (numbered according to FIAT conventions), shape
           (nfacet, ).
        """

    @abstractmethod
    def physical_points(self, point_set, entity=None):
        """Maps reference element points to GEM for the physical coordinates

        :arg point_set: A point_set on the reference cell to push forward to physical space.
        :arg entity: Reference cell entity on which the point set is
                     defined (for example if it is a point set on a facet).
        :returns: a GEM expression for the physical locations of the
                  points, shape (gdim, ) with free indices of the point_set.
        """

    @abstractmethod
    def physical_vertices(self):
        """Physical locations of the cell vertices.

        :returns: a GEM expression for the physical vertices, shape
                (gdim, )."""


zero = gem.Zero()
one = gem.Literal(1.0)


def identity(*shape):
    V = numpy.eye(*shape, dtype=object)
    for multiindex in numpy.ndindex(V.shape):
        V[multiindex] = zero if V[multiindex] == 0 else one
    return V



# Copyright (C) 2008 Robert C. Kirby (Texas Tech University)
# Modified by Andrew T. T. McRae (Imperial College London)
#
# This file is part of FIAT (https://www.fenicsproject.org)
#
# SPDX-License-Identifier:    LGPL-3.0-or-later

from FIAT import finite_element, polynomial_set, dual_set, functional
from FIAT.orientation_utils import make_entity_permutations_simplex
from FIAT.barycentric_interpolation import LagrangePolynomialSet, get_lagrange_points
from FIAT.reference_element import LINE
from FIAT.check_format_variant import parse_lagrange_variant


class LagrangeDualSet(dual_set.DualSet):
    """The dual basis for Lagrange elements.  This class works for
    simplicial complexes of any dimension.  Nodes are point evaluation at
    recursively-defined points.

    :arg ref_el: The simplicial complex.
    :arg degree: The polynomial degree.
    :arg point_variant: The point distribution variant passed on to recursivenodes.
    :arg sort_entities: A flag to sort entities by support vertex ids.
                        If false then entities are sorted first by dimension and then by
                        entity id. The DOFs are always sorted by the entity ordering
                        and then lexicographically by lattice multiindex.
    """
    def __init__(self, ref_el, degree, point_variant="equispaced", sort_entities=False):
        nodes = []
        entity_ids = {}
        entity_permutations = {}
        top = ref_el.get_topology()
        for dim in sorted(top):
            entity_ids[dim] = {}
            entity_permutations[dim] = {}
            perms = {0: [0]} if dim == 0 else make_entity_permutations_simplex(dim, degree - dim)
            for entity in sorted(top[dim]):
                entity_permutations[dim][entity] = perms

        entities = [(dim, entity) for dim in sorted(top) for entity in sorted(top[dim])]
        if sort_entities:
            # sort the entities by support vertex ids
            support = [top[dim][entity] for dim, entity in entities]
            entities = [entity for verts, entity in sorted(zip(support, entities))]

        # make nodes by getting points
        # need to do this entity-by-entity
        for dim, entity in entities:
            cur = len(nodes)
            pts_cur = ref_el.make_points(dim, entity, degree, variant=point_variant)
            nodes.extend(functional.PointEvaluation(ref_el, x) for x in pts_cur)
            entity_ids[dim][entity] = list(range(cur, len(nodes)))
        super().__init__(nodes, ref_el, entity_ids, entity_permutations)


class Lagrange(finite_element.CiarletElement):
    """The Lagrange finite element.

    :arg ref_el:  The reference element, which could be a standard FIAT simplex or a split complex
    :arg degree:  The polynomial degree
    :arg variant: A comma-separated string that may specify the type of point distribution
                  and the splitting strategy if a macro element is desired.
                  Either option may be omitted.  The default point type is equispaced
                  and the default splitting strategy is None.
                  Example: variant='gll' gives a standard unsplit point distribution with
                              spectral points.
                           variant='equispaced,Iso(2)' with degree=1 gives the P2:P1 iso element.
                           variant='Alfeld' can be used to obtain a barycentrically refined
                              macroelement for Scott-Vogelius.
    :arg sort_entities: A flag to sort entities by support vertex ids.
                        If false then entities are sorted first by dimension and then by
                        entity id. The DOFs are always sorted by the entity ordering
                        and then lexicographically by lattice multiindex.
    """
    def __init__(self, ref_el, degree, variant="equispaced", sort_entities=False):
        splitting, point_variant = parse_lagrange_variant(variant)
        if splitting is not None:
            ref_el = splitting(ref_el)
        dual = LagrangeDualSet(ref_el, degree, point_variant=point_variant, sort_entities=sort_entities)
        if ref_el.shape == LINE:
            # In 1D we can use the primal basis as the expansion set,
            # avoiding any round-off coming from a basis transformation
            points = get_lagrange_points(dual)
            poly_set = LagrangePolynomialSet(ref_el, points)
        else:
            poly_variant = "bubble" if ref_el.is_macrocell() else None
            poly_set = polynomial_set.ONPolynomialSet(ref_el, degree, variant=poly_variant)
        formdegree = 0  # 0-form
        super().__init__(poly_set, dual, degree, formdegree)



"""Utility functions for decomposing Concatenate nodes.

The exported functions are flatten and unconcatenate.
- flatten: destroys the structure preserved within Concatenate nodes,
           essentially reducing FInAT provided tabulations to what
           FIAT could have provided, so old code can continue to work.
- unconcatenate: split up (variable, expression) pairs along
                 Concatenate nodes, thus recovering the structure
                 within them, yet eliminating the Concatenate nodes.

Let us see an example on unconcatenate.  Let us consider the form

    div(v) * dx

where v is an RTCF7 test function.  This means that the assembled
local vector has 8 * 7 + 7 * 8 = 112 entries.  So the compilation of
the form starts with a single assignment pair [(v, e)].  v is now the
indexed return variable, something equivalent to

    Indexed(Variable('A', (112,)), (j,))

where j is the basis function index of the argument.  e is just a GEM
quadrature expression with j as its only free index.  This will
contain the tabulation of the RTCF7 element, which will cause
something like

    C_j := Indexed(Concatenate(A, B), (j,))

to appear as a subexpression in e.  unconcatenate splits e along C_j
into e_1 and e_2 such that

    e_1 := e /. C_j -> A_{ja1,ja2}, and
    e_2 := e /. C_j -> B_{jb1,jb2}.

The split indices ja1, ja2, jb1, and jb2 have extents 8, 7, 7, and 8
respectively (see the RTCF7 element construction above).  So the
result of unconcatenate will be the list of pairs

     [(v_1, e_2), (v_2, e_2)]

where v_1 is the first 56 entries of v, reshaped as an 8 x 7 matrix,
indexed with (ja1, ja2), and similarly, v_2 is the second 56 entries
of v, reshaped as a 7 x 8 matrix, indexed with (jb1, jb2).

The unconcatenated form allows for sum factorisation of tensor product
elements as usual.  This pair splitting is also applicable to
coefficient evaluation: take the local basis function coefficients as
the variable, the FInAT tabulation of the element as the expression,
and apply "matrix-vector multifunction" for each pair after
unconcatenation, and then add up the results.
"""

from functools import singledispatch
from itertools import chain

import numpy

from gem.node import Memoizer, reuse_if_untouched
from gem.gem import (ComponentTensor, Concatenate, FlexiblyIndexed,
                     Index, Indexed, Literal, Node, partial_indexed,
                     reshape, view)
from gem.optimise import remove_componenttensors
from gem.interpreter import evaluate


__all__ = ['flatten', 'unconcatenate']


def find_group(expressions):
    """Finds a full set of indexed Concatenate nodes with the same
    free index, if any such node exists.

    Pre-condition: ComponentTensor nodes surrounding Concatenate nodes
    must be removed.

    :arg expressions: a multi-root GEM expression DAG
    :returns: a list of GEM nodes, or None
    """
    free_indices = set().union(chain(*[e.free_indices for e in expressions]))

    # Result variables
    index = None
    nodes = []

    # Sui generis pre-order traversal so that we can avoid going
    # unnecessarily deep in the DAG.
    seen = set()
    lifo = []
    for root in expressions:
        if root not in seen:
            seen.add(root)
            lifo.append(root)

    while lifo:
        node = lifo.pop()
        if not free_indices.intersection(node.free_indices):
            continue

        if isinstance(node, Indexed):
            child, = node.children
            if isinstance(child, Concatenate):
                i, = node.multiindex
                assert i in free_indices
                if (index or i) == i:
                    index = i
                    nodes.append(node)
                    # Skip adding children
                    continue

        for child in reversed(node.children):
            if child not in seen:
                seen.add(child)
                lifo.append(child)

    return index and nodes


def split_variable(variable_ref, index, multiindices):
    """Splits a flexibly indexed variable along a concatenation index.

    :param variable_ref: flexibly indexed variable to split
    :param index: :py:class:`Concatenate` index to split along
    :param multiindices: one multiindex for each split variable

    :returns: generator of split indexed variables
    """
    assert isinstance(variable_ref, FlexiblyIndexed)
    other_indices = list(variable_ref.index_ordering())
    other_indices.remove(index)
    other_indices = tuple(other_indices)
    data = ComponentTensor(variable_ref, (index,) + other_indices)
    slices = [slice(None)] * len(other_indices)
    shapes = [(other_index.extent,) for other_index in other_indices]

    offset = 0
    for multiindex in multiindices:
        shape = tuple(index.extent for index in multiindex)
        size = numpy.prod(shape, dtype=int)
        slice_ = slice(offset, offset + size)
        offset += size

        sub_ref = Indexed(reshape(view(data, slice_, *slices),
                                  shape, *shapes),
                          multiindex + other_indices)
        sub_ref, = remove_componenttensors((sub_ref,))
        yield sub_ref


def _replace_node(node, self):
    """Replace subexpressions using a given mapping.

    :param node: root of expression
    :param self: function for recursive calls
    """
    assert isinstance(node, Node)
    if self.cut(node):
        return node
    try:
        return self.mapping[node]
    except KeyError:
        return reuse_if_untouched(node, self)


def replace_node(expression, mapping, cut=None):
    """Replace subexpressions using a given mapping.

    :param expression: a GEM expression
    :param mapping: a :py:class:`dict` containing the substitutions
    :param cut: cutting predicate; if returns true, it is assumed that
                no replacements would take place in the subexpression.
    """
    mapper = Memoizer(_replace_node)
    mapper.mapping = mapping
    mapper.cut = cut or (lambda node: False)
    return mapper(expression)


def _unconcatenate(cache, pairs):
    # Tail-call recursive core of unconcatenate.
    # Assumes that input has already been sanitised.
    concat_group = find_group([e for v, e in pairs])
    if concat_group is None:
        return pairs

    # Get the index split
    concat_ref = next(iter(concat_group))
    assert isinstance(concat_ref, Indexed)
    concat_expr, = concat_ref.children
    index, = concat_ref.multiindex
    assert isinstance(concat_expr, Concatenate)
    try:
        multiindices = cache[index]
    except KeyError:
        multiindices = tuple(tuple(Index(extent=d) for d in child.shape)
                             for child in concat_expr.children)
        cache[index] = multiindices

    def cut(node):
        """No need to rebuild expression of independent of the
        relevant concatenation index."""
        return index not in node.free_indices

    # Build Concatenate node replacement mappings
    mappings = [{} for i in range(len(multiindices))]
    for concat_ref in concat_group:
        concat_expr, = concat_ref.children
        for i in range(len(multiindices)):
            sub_ref = Indexed(concat_expr.children[i], multiindices[i])
            sub_ref, = remove_componenttensors((sub_ref,))
            mappings[i][concat_ref] = sub_ref

    # Finally, split assignment pairs
    split_pairs = []
    for var, expr in pairs:
        if index not in var.free_indices:
            split_pairs.append((var, expr))
        else:
            for v, m in zip(split_variable(var, index, multiindices), mappings):
                split_pairs.append((v, replace_node(expr, m, cut)))

    # Run again, there may be other Concatenate groups
    return _unconcatenate(cache, split_pairs)


def unconcatenate(pairs, cache=None):
    """Splits a list of (indexed variable, expression) pairs along
    :py:class:`Concatenate` nodes embedded in the expressions.

    :param pairs: list of (indexed variable, expression) pairs
    :param cache: index splitting cache :py:class:`dict` (optional)

    :returns: list of (indexed variable, expression) pairs
    """
    # Set up cache
    if cache is None:
        cache = {}

    # Eliminate index renaming due to ComponentTensor nodes
    exprs = remove_componenttensors([e for v, e in pairs])
    pairs = [(v, e) for (v, _), e in zip(pairs, exprs)]

    return _unconcatenate(cache, pairs)


@singledispatch
def _flatten(node, self):
    """Replace Concatenate nodes with Literal nodes.

    :arg node: root of the expression
    :arg self: function for recursive calls
    """
    raise AssertionError("cannot handle type %s" % type(node))


_flatten.register(Node)(reuse_if_untouched)


@_flatten.register(Concatenate)
def _flatten_concatenate(node, self):
    result, = evaluate([node])
    return partial_indexed(Literal(result.arr), result.fids)


def flatten(expressions):
    """Flatten Concatenate nodes, and destroy the structure they express.

    :arg expressions: a multi-root expression DAG
    """
    mapper = Memoizer(_flatten)
    return list(map(mapper, expressions))
